/*
   common.h -- PIKT common header file for all binaries

   $Header: /cvs/PIKT/pikt/src/lib/common.h,v 1.76 2002/08/23 16:42:05 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <sys/param.h>
#include <sys/resource.h>

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>		/* varargs.h? When? */
#include <stdlib.h>

#if STDC_HEADERS
# include <string.h>
#else
# ifdef HAVE_STRING_H
#  include <string.h>
# endif
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#include <syslog.h>
#include <time.h>
#include <unistd.h>

#ifdef WITH_REGEX 
#include <regex.h>
#else
#include <rxposix.h>
#endif

#include "errid.h"

#ifndef S_IFSOCK
# define S_IFSOCK	0140000		/* socket */
#endif
#ifndef S_ISSOCK
# define S_ISSOCK(x)	((x & S_IFMT) == S_IFSOCK)
#endif

#define MAXLINE		512	/* max length of input data line, or max
				   line length in various other contexts
				 */

#define MAXCMD		MAXLINE	/* max length of scancmd, also for exec'ed
				   commands; a cmd longer than this will
				   dump core; we don't error check this, but
				   make this generously long
				 */

#define MAXKEY		MAXLINE	/* max length of key value; we set this to
				   MAXLINE, which is way generous
				 */

#define MAXMCRBUF	MAXLINE	/* max length of various macro buffers */


#define MAXBUF		65536	/* max length of function buffer string */

#ifndef MAXPATH
#define MAXPATH		256	/* max length of file path */
#endif

#define MAXIDENT	256	/* max length of ident */

#define MAXIFS		64	/* applies to "if", "#if", "#if()", "$if()",
				   also to "#include"
				 */

#define MAXMCRPRMS	32	/* max number of macro parameters;
				   do not change this!
				 */

#define MAXWHILES	MAXIFS
#define MAXREPEATS	MAXIFS
#define MAXFORS		MAXIFS
#define MAXSWITCHS	MAXIFS
#define MAXLOOPS	3*MAXIFS

#define MAXARY		MAXIFS

#define MAXVARS		255	/* max number of user-defined var names */

#ifndef MAXUID
#define MAXUID		65534
#endif
#ifndef MAXGID
#define MAXGID		65534
#endif

#define MAXTIMINGS	16		/* max number of alert timings */

#define MALLOC_TIME_TO_SLEEP	60	/* number of seconds to pause if
					   exhaustion of virtual memory
					   causes any of the malloc() family
					   of functions to return NULL
					 */

#define BIGENDIAN		1
#define LITTLEENDIAN		2

/* add to this list of encryption algorithms as necessary */
#define NO_ENCRYPTION		0
#define BLOWFISH_ENCRYPTION	1
#define MAX_ENCRYPTION_TYPE	BLOWFISH_ENCRYPTION

#define BF_MAXKEYBYTES		56	/* max blowfish key length, 448 bits */
#define MAXKEYBYTES		BF_MAXKEYBYTES

#define EXECCMDOUT_LIMIT	1000000		/* default max number of
						   characters allowed for
						   execute_commandwait()
						   output; can be overridden
						   in PIKT.conf
						 */

#define RPC_TIMEOUT		60	/* number of seconds piktc will wait
					   for return of execute_command()
					   or execute_alert() call before
					   timing out
					 */

#define RPC_RETRY_TIMEOUT	RPC_TIMEOUT*2
					/* number of seconds piktc will wait
					   for return of execute_command()
					   or execute_alert() call before
					   retrying the request; by default,
					   never retry
					 */

#define RPCINFO_RETRIES		1	/* number of rpcinfo retries (after
					   a 'timed out' result) in
					   chkpiktc_svc() beyond which we
					   just give up
					 */

#define RESTART_PIKTC_SVC_WAIT	5	/* when doing a restart_piktc_svc
					   (piktc -R), number of seconds piktc
					   waits before sending second
					   piktc_svc_pid() to verify restart
					 */

#define PIKTNEVER	"0 0 0 0 0 0"	/* signals to readconf() that this
					   alert is never invoked by piktd
					 */

#define ERROR		-1
#define NONE		-2
#define DENIED		-3
#define UNAUTHORIZED	-4
#define IGNORE		-5

#define NIL		NONE	/* for value unknown */

#ifndef FALSE
#define FALSE		0
#endif
#ifndef TRUE
#define TRUE		1
#endif

/* services */
#define KILL_PIKTD_SERVICE		0x00000001
#define KILL_PIKTC_SVC_SERVICE		0x00000002
#define KILL_SERVICE			KILL_PIKTD_SERVICE | \
					KILL_PIKTC_SVC_SERVICE

#define RESTART_PIKTD_SERVICE		0x00000004
#define RESTART_PIKTC_SVC_SERVICE	0x00000008
#define RESTART_SERVICE			RESTART_PIKTD_SERVICE | \
					RESTART_PIKTC_SVC_SERVICE

#define CALL_BACK_SERVICE		0x00000010
#define GET_KEY_SERVICE			0x00000020

#define EXECUTE_COMMANDS_SERVICE	0x00000040
#define EXECUTE_ALERTS_SERVICE		0x00000080
#define EXECUTE_SERVICE			EXECUTE_COMMANDS_SERVICE | \
					EXECUTE_ALERTS_SERVICE

#define DISABLE_ALERTS_SERVICE		0x00000100
#define ENABLE_ALERTS_SERVICE		0x00000200

#define SHOW_ALERTS_STATUS_SERVICE	0x00000400
/* reserved:				0x00000800 */

#define INSTALL_ALERTS_SERVICE		0x00001000
#define INSTALL_PROGRAMS_SERVICE	0x00002000
#define INSTALL_FILES_SERVICE		0x00004000
#define INSTALL_OBJECTS_SERVICE		0x00008000
#define INSTALL_SERVICE			INSTALL_ALERTS_SERVICE | \
					INSTALL_PROGRAMS_SERVICE | \
					INSTALL_FILES_SERVICE | \
					INSTALL_OBJECTS_SERVICE

#define FETCH_ALERTS_SERVICE		0x00010000
#define FETCH_PROGRAMS_SERVICE		0x00020000
#define FETCH_FILES_SERVICE		0x00040000
#define FETCH_OBJECTS_SERVICE		0x00080000
#define FETCH_SERVICE			FETCH_ALERTS_SERVICE | \
					FETCH_PROGRAMS_SERVICE | \
					FETCH_FILES_SERVICE | \
					FETCH_OBJECTS_SERVICE

#define CKSUM_ALERTS_SERVICE		0x00100000
#define CKSUM_PROGRAMS_SERVICE		0x00200000
#define CKSUM_FILES_SERVICE		0x00400000
#define CKSUM_OBJECTS_SERVICE		0x00800000
#define CKSUM_SERVICE			CKSUM_ALERTS_SERVICE | \
					CKSUM_PROGRAMS_SERVICE | \
					CKSUM_FILES_SERVICE | \
					CKSUM_OBJECTS_SERVICE

#define DELETE_ALERTS_SERVICE		0x01000000
#define DELETE_PROGRAMS_SERVICE		0x02000000
#define DELETE_FILES_SERVICE		0x04000000
#define DELETE_OBJECTS_SERVICE		0x08000000
#define DELETE_SERVICE			DELETE_ALERTS_SERVICE | \
					DELETE_PROGRAMS_SERVICE | \
					DELETE_FILES_SERVICE | \
					DELETE_OBJECTS_SERVICE

#define LIST_ALERTS_SERVICE		0x10000000
#define LIST_PROGRAMS_SERVICE		0x20000000
#define LIST_FILES_SERVICE		0x40000000
#define LIST_OBJECTS_SERVICE		0x80000000
#define LIST_SERVICE			LIST_ALERTS_SERVICE | \
					LIST_PROGRAMS_SERVICE | \
					LIST_FILES_SERVICE | \
					LIST_OBJECTS_SERVICE

#define ALL_SERVICES			0xffffffff

#define IS_CALL_BACK			(services & CALL_BACK_SERVICE)

/* dat type */
#define STRING		0
#define NUMBER		1
#define FILEHANDLE	2
#define PROCHANDLE	3

/* statement type */
#define DO		1
#define IF		2
#define ELIF		3
#define ELSE		4
#define ENDIF		5
#define WHILE		6
#define ENDWHILE	7
#define WHILEAUX	107
#define REPEAT		8
#define UNTIL		9
#define REPEATAUX	109
#define FOR		10
#define ENDFOR		11
#define FORAUX		111
#define BREAK		12
#define CONTINUE	13
#define SWITCH		14
#define CASE		15
#define DEFAULT		16
#define BREAKSWITCH	17
#define ENDSWITCH	18

#define AGAIN		19
#define LEAVE		20
#define REDO		21
#define NEXT		22
#define SKIP		23
#define LAST		24
#define PAUSE		25
#define QUIT		26
#define DIE		27

#define OUTPUT		28
#define MAIL		29
#define PRINT		30
#define LOCLOG		31
#define PIKTLOG		32
#define SYSLOG		33

#define SET		34
#define UNSET		35
#define EXEC		36
#define EXECWAIT	37

/* procid type */
#define PROCID		0
#define PPROCID		1

#define SHELL		"/bin/sh"

#define ISVAR(c)	(strchr("$#%@",c) && (isalnum((int)*(cp+1)) || (*(cp+1) == '\[')) && ! escape)

#define CLEANCHAR	46	/* '.', the "clean" char we substitute for
				   negative signed char values when
				   "sanitizing" log strings
				 */

/* choose obscure, infrequently used special chars for SQTPROXY, DQTPROXY &
   PNDPROXY
 */
#define SQTPROXY	14	/* (^n, shift-out); single-quote stand-in */
#define DQTPROXY	15	/* (^o, shift-in);  double-quote stand-in */
#define PNDPROXY	16	/* (^p, data link escape);  # stand-in */

#define LP	40	/* left parenthesis */
#define RP	41	/* right parenthesis */
#define LB	91	/* left bracket */
#define RB	93	/* right bracket */

#define foreach(a,b)	for (a=b; a; a=a->next)
#define maxval(a,b)	((a) >= (b) ? (a) : (b))
#define minval(a,b)	((a) <= (b) ? (a) : (b))
#define odd(a)		(((a) % 2) ? TRUE : FALSE)
#define even(a)		(((a) % 2) ? FALSE : TRUE)

#define FREE(a)		if (a) free(a)
#define FREEVOID(a)	if (a) free(a); a = (void *) NULL

#define BF_ROUNDS 16	/* better to leave unchanged */

/* the first char for each of these #define's must be uppercase, else change
   src/piktc/rpcsvc.c's getlocdir() accordingly; the length of each of the
   following should be same as DIRLEN
 */
#define HOMDIR		"HOMDIR"
#define BINDIR		"BINDIR"
#define ETCDIR		"ETCDIR"
#define LOKDIR		"LOKDIR"
#define LIBDIR		"LIBDIR"
#define VARDIR		"VARDIR"
#define ALTDIR		"ALTDIR"
#define PRGDIR		"PRGDIR"
#define OBJDIR		"OBJDIR"
#define CFGDIR		"CFGDIR"
#define DIFDIR		"DIFDIR"
#define STGDIR		"STGDIR"
#define HSTDIR		"HSTDIR"
#define LOGDIR		"LOGDIR"

#define DIRLEN		6

#define BAKEXT		"piktbak"

typedef struct {
    unsigned long s[4][256];
    unsigned long p[BF_ROUNDS+2];
} blf_ctx;

extern blf_ctx bc;	/* blowfish context */

extern int endian;	/* the endianness of this system */

extern char **argvcurr;	/* current command-line argument */
extern char **argvlast;	/* last command-line argument */

extern int scanpass;	/* is this the 1st, 2nd, 3rd, or ... scan pass? */

extern int lineno;	/* line number in config or script file */

extern int bailout;	/* general, all-purpose abort signal */
extern int bypass;	/* limited, special-purpose abort signal */

extern int errno;

extern int exstat;
extern const int exit_status[];
extern const char *error_level[];

/* command-line options */
extern int isbackup,
           ischeck,
           isdebug,
           isdelete,
           iscksum,
           isdiff,
           isdoublecheck,
           isdisable,
           isenable,
           isexec,
           isexecwait,
	   isexpand_macros,
           ishelp,
           isinstall,
           isupdate_includes,
           iskill_piktd,
           iskill_piktc_svc,
           islist_elements,
           islist_files,
           islist_items,   
           isinstall_piktconf,
           isrestart_piktd,
           isrestart_piktc_svc,
           isshow,
	   isshow_version,
           istest,
           isverbose;

/* dir/file paths */
extern char *homdir,		/* PIKT home directory */
            *bindir,		/* binaries directory */
            *etcdir,		/* admin directory */
            *lokdir,		/* lock file directory */
            *libdir,		/* lib directory */
            *vardir,		/* var directory */
            *altdir,		/* alerts directory */
            *prgdir,		/* programs directory */
            *objdir,		/* objects directory */
            *cfgdir,		/* configs directory */
            *difdir,		/* diffing directory */
            *stgdir,		/* staging directory */
            *hstdir,		/* histories directory */
            *logdir,		/* log directory */
            *PIKTconf,		/* PIKT conf file */
            *keysconf,		/* keys conf file */
            *pikt,		/* pikt program */
            *piktd,		/* piktd program */
            *piktc,		/* piktc program */
            *piktc_svc,		/* piktc_svc program */
            *pikts,		/* pikts program */
            *piktx,		/* piktx program */
            *piktdconf,		/* piktd conf file */
            *piktdconftmp,	/* piktd conf temp file */
            *piktdlock,		/* piktd lock file */
            *piktdlog,		/* piktd log file */
            *piktdalt,		/* piktd alerts file */
            *piktclock,		/* piktc lock file */
            *piktclog,		/* piktc log file */
            *piktc_svclock,	/* piktc_svc lock file */
            *piktc_svclog;	/* piktc_svc log file */

extern char *arpcmd,		/* arp command path */
            *cksumcmd,		/* cksum command path */
            *cpcmd,		/* cp command path */
            *diffcmd,		/* diff command path */
            *egrepcmd,		/* egrep command path */
            *filecmd,		/* file command path */
            *pscmd,		/* ps command path */
            *rmcmd,		/* rm command path */
            *rpcinfocmd,	/* rpcinfo command path */
            *scpcmd,		/* scp command path */
            *sumcmd,		/* sum command path */
            *tailcmd,		/* tail command path */
            *trcmd;		/* tr command path */

extern char *bakext;		/* backup extension for installed files */

/* PIKT.conf options */
extern AUTHUNIX_CREATE_UID_T authuid;	/* piktc/piktc_svc service
					   request & PIKT.conf uid
					 */
extern AUTHUNIX_CREATE_GID_T authgid;	/* piktc/piktc_svc service
					   request & PIKT.conf gid
					 */
extern char *master;		/* hostname of master control system */
extern char *domain;		/* local network domain, or "domain" */
extern char *accessmode;	/* "local" or "global" */
extern char *master_address;	/* ip address of master control system */
extern char *master_ethernet_address;	/* ip address of master
					   control system
					 */
extern int isauth_by_master;	/* do piktc request authentication
				   by master hostname?
				 */
extern int isauth_by_master_address;	/* do piktc request authentication
				   	   by master ip address?
					 */
extern int isauth_by_master_ethernet_address;	/* do piktc request
						   authentication by master
						   ethernet address?
						 */
extern int data_encryption_type;	/* 0 is no encryption of data */
extern int auth_encryption_type;	/* 0 is no encryption of key */
extern char *pvtkey;			/* private encryption key */
extern unsigned int pvtkeylen;		/* length of pvtkey */
extern int rpc_timeout;		/* number of seconds piktc will wait
				   for return of execute_command() or
				   execute_alert() call before timing out
				 */
extern int rpc_retry_timeout;	/* number of seconds piktc will wait
				   for return of execute_command() or
				   execute_alert() call before retrying
				   the request
				 */
extern int rpcinfo_retries;	/* number of rpcinfo retries (after
				   a 'timed out' result) in
				   chkpiktc_svc() beyond which we
				   just give up
				 */
extern int restart_piktc_svc_wait;	/* when doing a restart_piktc_svc
					   (piktc -R), number of seconds piktc
					   waits before sending second
					   piktc_svc_pid() to verify restart
					 */
extern int isverboselog;	/* increase the verbosity of logging */
extern int isautosyslog;	/* do we automatically log to syslog also? */
extern int istcponly;		/* use tcp only, no udp, in piktc-to-piktc_svc
				   communications
				 */
extern int execcmdout_limit;	/* when doing a 'piktc -x', the maximum
				   length in chars of returned command
				   output
				 */

extern int PIKT_conf_gid;	/* is gid set in PIKT.conf? */

extern unsigned long services;		/* piktc-to-piktc_svc access
					   privileges
					 */

extern unsigned long confparsrv;	/* the PIKT.conf service */
extern char **confparstrp;		/* the PIKT.conf string parameter */
extern int *confparintp;		/* the PIKT.conf int parameter */

extern char hostname[];		/* hostname of system that piktc and/or
				   piktc_svc runs on; or name of alert
				   host
			 	 */
extern char domainname[];	/* domainname of system that piktc and/or
				   piktc_svc runs on; or domainname of
				   alert host
			 	 */

extern int ismaster;		/* is this host the master? */

extern const char *leveltypes[];	/* EMERG/ALERT/CRIT/... */

extern int syslog_facility;

extern unsigned int fork_delay;

extern time_t time_start;		/* program start time */
extern pid_t prgpid;			/* program pid */
extern pid_t prgppid;			/* program ppid */
extern char *prgnam;			/* program name */

extern const signed char nil[];		/* unknown val */
extern const signed char err[];		/* error val */
extern const char nilstr[];		/* unknown val */
extern const char errstr[];		/* error val */

extern const signed char dquote[];	/* double quote */
extern const signed char squote[];	/* single quote */
extern const char space[];

extern const char keywords_sections[];			/* init, begin,
							   rule, end
							*/
extern const char keywords_statements[];		/* all other
							   keywords
							*/
extern const char keywords_statements_block_begin[];	/* keywords beginning
							   a statement block
							 */
extern const char keywords_statements_block_end[];	/* keywords ending
							   a statement block
							 */

extern const char *tabs[];			/* formatting tab strings */

extern const char nullstring[];			/* zero-length string */

/* init functions */
extern void init(int, char **, void (*)(void), const char *);
extern void getconf(char *);
extern void setstr(char *, char **, char *);
extern void setint(char *, int *, int);
extern void setsrv(char *, unsigned long, int);
extern int hostmatch(char *, char *);
extern char *getlocdir(char *);
extern int isfile(char *, int);
extern int daemon_init(const int);

/* lexer and parser functions */
extern int confparse(void);
extern int conflex(void);
extern void conferror(const char *);

/* lex functions */
extern int condparen(char *);
extern int leadstr(char *);
extern int decompose(signed char *, signed char *);
extern int getexecunput(char **, char *, int, int, int, int);

/* text functions */
extern char *stripesc(char *);
extern char *addesc(char *);
extern int dequote(signed char *);
extern int requote(signed char *);

/* mem functions */
extern void *mymalloc(size_t, const char *);
extern void *mycalloc(size_t, size_t, const char *);
extern signed char *mystrdup(const signed char *, const char *);

/* pid functions */
extern pid_t procid(char *, int);

/* lock functions */
extern int islock(char *);
extern int mklock(char *, char *, int, int);
extern int rmlock(char *, int, int);
extern int rmlocks(int, int);

/* logging functions */
extern void logreturn(const int, const int, const int, const char *, const int, const char *, const int, const char *, ...);
extern void logexit(const int, const int, const int, const char *, const int, const char *, const int, const char *, ...);

/* signal functions */
extern void sig_init(void);

/* ipc functions */
extern void ipcopen(int, char *, int);
extern void ipcclose(int, char *, int);
extern void tell_parent(int, char *, int);
extern void tell_child(int, char *, int);
extern void wait_parent(int, char *, int);
extern void wait_child(int, char *, int);

/* network functions */
extern char *get_ethaddr(char *);
extern int ethaddr_match(char *, char *);

/* time functions */
extern char *get_timenow(void);

/* checksum functions */
extern char *checksumfil(int, signed char *);
extern char *checksumstr(int, signed char *);
extern int csstrcmp(char *, char *);

/* crypt functions */
extern int bf_setkey(blf_ctx *, char *, unsigned);
extern void bf_encrypt(blf_ctx *, char *, unsigned int, char **, unsigned int *);
extern void bf_decrypt(blf_ctx *, char *, unsigned int, char **, unsigned int *);
extern void fix_endian(char *, unsigned int);
extern char *sanitize (char *, unsigned int);

/* version functions */
extern void show_version(const char *);
