%{

/*
   cmd.l -- PIKT piktc command-line lexer

   $Header: /cvs/PIKT/pikt/src/piktc/cmd.l,v 1.47 2002/08/23 16:55:29 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"
#include "cmd.tab.h"

char _cmdlex_[] = "cmdlex";

#define RETURN(token) cmdlval.sval = mystrdup(yytext, _cmdlex_); return token

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size)	{ \
		signed char c = cmdinput(); \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

#define CHKCONFLICT(A, B, FL)	if (FL) {\
					logreturn(OUTPUT, LOG_ERR, ERRID_CONFLICTING_OPTIONS, __FILE__, __LINE__, _cmdlex_, FALSE, "conflicting options: %s %s", A, B); \
					bailout = TRUE; \
				}

unsigned aoffset = 0;

char errmsg[MAXLINE+1];

char ALL[] = "+A all +P all +F all +O all +H all ";
int ii;

void cmderror(const char *);
signed int cmdinput(void);
void wrapstr(char **);
void unwrapstr(char **);

%}

W		[ \t]
nW		[^ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%x mS pS DS CS AS PS FS OS HS

%option noyywrap

%%

<INITIAL,DS,AS,PS,FS,OS,HS>"-"	{ BEGIN(mS);
				  return _MINUS;
				}
<INITIAL,DS,AS,PS,FS,OS,HS>"+"	{ BEGIN(pS);
				  return _PLUS;
				}

<mS,pS>(A|S)f?			{ BEGIN(AS);
				  if (scanpass == 1) {
					CHKCONFLICT((yytext[0] == 'A' ? "+/-A" : "+/-S"), "-c", ischeck);
					CHKCONFLICT((yytext[0] == 'A' ? "+/-A" : "+/-S"), "-C", isdoublecheck);
					CHKCONFLICT((yytext[0] == 'A' ? "+/-A" : "+/-S"), "-h", ishelp);
					CHKCONFLICT((yytext[0] == 'A' ? "+/-A" : "+/-S"), "-V", isshow_version);
					if (yytext[1] == 'f')
						stripalm = FALSE;
					rcfbs |= (1L << ALMCFG);
					rcfbs |= (1L << ALTCFG);
					havecmditems = TRUE;
				  }
				  return _ALTOPT;  /* _SCTOPT */
				}

<mS>b				{ if (scanpass == 1) {
					isbackup = TRUE;
					CHKCONFLICT("-b", "-c", ischeck);
					CHKCONFLICT("-b", "-C", isdoublecheck);
					CHKCONFLICT("-b", "-d", isdisable);
					CHKCONFLICT("-b", "-E", islist_elements);
					CHKCONFLICT("-b", "-f", isdiff);
					CHKCONFLICT("-b", "-h", ishelp);
					CHKCONFLICT("-b", "-k", iskill_piktd);
					CHKCONFLICT("-b", "-K", iskill_piktc_svc);
					CHKCONFLICT("-b", "-l", islist_files);
					CHKCONFLICT("-b", "-L", islist_items);
					CHKCONFLICT("-b", "-m", iscksum);
					CHKCONFLICT("-b", "-M", isexpand_macros);
					CHKCONFLICT("-b", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-b", "-s", isshow);
					CHKCONFLICT("-b", "-V", isshow_version);
					CHKCONFLICT("-b", "-x", isexecwait);
					CHKCONFLICT("-b", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>c|C				{ if (yytext[0] == 'c') {
					if (scanpass == 1) {
						CHKCONFLICT("-c", "-C", isdoublecheck);
						ischeck = TRUE;
					}
				  } else {
					if (scanpass == 1) {
						CHKCONFLICT("-C", "-c", ischeck);
						isdoublecheck = TRUE;
					}
				  }
				  if (scanpass == 1) {
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-b", isbackup);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-d", isdisable);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-e", isenable);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-E", islist_elements);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-f", isdiff);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-h", ishelp);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-i", isinstall);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-k", iskill_piktd);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-K", iskill_piktc_svc);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-l", islist_files);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-L", islist_items);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-m", iscksum);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-M", isexpand_macros);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-p", isinstall_piktconf);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-r", isrestart_piktd);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-R", isrestart_piktc_svc);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-s", isshow);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-t", isdelete);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-V", isshow_version);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-x", isexecwait);
					CHKCONFLICT((yytext[0] == 'c' ? "-c" : "-C"), "-X", isexec);
				  }
				  return _OPT;
				}

<pS>C|M				{ BEGIN(CS);
				  if (scanpass == 1) {
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-b", isbackup);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-c", ischeck);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-C", isdoublecheck);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-d", isdisable);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-e", isenable);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-E", islist_elements);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-f", isdiff);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-h", ishelp);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-i", isinstall);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-k", iskill_piktd);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-K", iskill_piktc_svc);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-l", islist_files);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-L", islist_items);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-m", iscksum);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-p", isinstall_piktconf);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-r", isrestart_piktd);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-R", isrestart_piktc_svc);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-s", isshow);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-t", isdelete);
					CHKCONFLICT((yytext[0] == 'C' ? "+C" : "+M"), "-V", isshow_version);
					havecmditems = TRUE;
				  }
				  return _CMDOPT;
				}

<mS>d				{ if (scanpass == 1) {
					isdisable = TRUE;
					CHKCONFLICT("-d", "-b", isbackup);
					CHKCONFLICT("-d", "-c", ischeck);
					CHKCONFLICT("-d", "-C", isdoublecheck);
					CHKCONFLICT("-d", "-e", isenable);
					CHKCONFLICT("-d", "-E", islist_elements);
					CHKCONFLICT("-d", "-f", isdiff);
					CHKCONFLICT("-d", "-h", ishelp);
					CHKCONFLICT("-d", "-i", isinstall);
					CHKCONFLICT("-d", "-K", iskill_piktc_svc);
					CHKCONFLICT("-d", "-l", islist_files);
					CHKCONFLICT("-d", "-L", islist_items);
					CHKCONFLICT("-d", "-m", iscksum);
					CHKCONFLICT("-d", "-M", isexpand_macros);
					CHKCONFLICT("-d", "-p", isinstall_piktconf);
					CHKCONFLICT("-d", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-d", "-s", isshow);
					CHKCONFLICT("-d", "-V", isshow_version);
					CHKCONFLICT("-d", "-x", isexecwait);
					CHKCONFLICT("-d", "-X", isexec);
				  }
				  return _OPT;
				}

<mS,pS>D			{ BEGIN(DS);
				  if (scanpass == 1) {
					CHKCONFLICT("+/-D", "-h", ishelp);
					CHKCONFLICT("+/-D", "-V", isshow_version);
					rcfbs |= (1L << DEFCFG);
				  }
				  return _DEFOPT;
				}

<mS>e				{ if (scanpass == 1) {
					isenable = TRUE;
					CHKCONFLICT("-e", "-c", ischeck);
					CHKCONFLICT("-e", "-C", isdoublecheck);
					CHKCONFLICT("-e", "-d", isdisable);
					CHKCONFLICT("-e", "-E", islist_elements);
					CHKCONFLICT("-e", "-f", isdiff);
					CHKCONFLICT("-e", "-h", ishelp);
					CHKCONFLICT("-e", "-K", iskill_piktc_svc);
					CHKCONFLICT("-e", "-l", islist_files);
					CHKCONFLICT("-e", "-L", islist_items);
					CHKCONFLICT("-e", "-m", iscksum);
					CHKCONFLICT("-e", "-M", isexpand_macros);
					CHKCONFLICT("-e", "-p", isinstall_piktconf);
					CHKCONFLICT("-e", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-e", "-s", isshow);
					CHKCONFLICT("-e", "-t", isdelete);
					CHKCONFLICT("-e", "-V", isshow_version);
					CHKCONFLICT("-e", "-x", isexecwait);
					CHKCONFLICT("-e", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>E				{ if (scanpass == 1) {
					islist_elements = TRUE;
					CHKCONFLICT("-E", "-b", isbackup);
					CHKCONFLICT("-E", "-c", ischeck);
					CHKCONFLICT("-E", "-C", isdoublecheck);
					CHKCONFLICT("-E", "-d", isdisable);
					CHKCONFLICT("-E", "-e", isenable);
					CHKCONFLICT("-E", "-f", isdiff);
					CHKCONFLICT("-E", "-h", ishelp);
					CHKCONFLICT("-E", "-i", isinstall);
					CHKCONFLICT("-E", "-k", iskill_piktd);
					CHKCONFLICT("-E", "-K", iskill_piktc_svc);
					CHKCONFLICT("-E", "-l", islist_files);
					CHKCONFLICT("-E", "-m", iscksum);
					CHKCONFLICT("-E", "-M", isexpand_macros);
					CHKCONFLICT("-E", "-p", isinstall_piktconf);
					CHKCONFLICT("-E", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-E", "-s", isshow);
					CHKCONFLICT("-E", "-V", isshow_version);
					CHKCONFLICT("-E", "-x", isexecwait);
					CHKCONFLICT("-E", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>f				{ if (scanpass == 1) {
					isdiff = TRUE;
					CHKCONFLICT("-f", "-b", isbackup);
					CHKCONFLICT("-f", "-c", ischeck);
					CHKCONFLICT("-f", "-C", isdoublecheck);
					CHKCONFLICT("-f", "-d", isdisable);
					CHKCONFLICT("-f", "-e", isenable);
					CHKCONFLICT("-f", "-E", islist_elements);
					CHKCONFLICT("-f", "-h", ishelp);
					CHKCONFLICT("-f", "-i", isinstall);
					CHKCONFLICT("-f", "-k", iskill_piktd);
					CHKCONFLICT("-f", "-K", iskill_piktc_svc);
					CHKCONFLICT("-f", "-l", islist_files);
					CHKCONFLICT("-f", "-L", islist_items);
					CHKCONFLICT("-f", "-m", iscksum);
					CHKCONFLICT("-f", "-M", isexpand_macros);
					CHKCONFLICT("-f", "-p", isinstall_piktconf);
					CHKCONFLICT("-f", "-r", isrestart_piktd);
					CHKCONFLICT("-f", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-f", "-s", isshow);
					CHKCONFLICT("-f", "-t", isdelete);
					CHKCONFLICT("-f", "-V", isshow_version);
					CHKCONFLICT("-f", "-x", isexecwait);
					CHKCONFLICT("-f", "-X", isexec);
				  }
				  return _OPT;
				}

<mS,pS>(F)f?			{ BEGIN(FS);
				  if (scanpass == 1) {
					CHKCONFLICT("+/-F", "-c", ischeck);
					CHKCONFLICT("+/-F", "-C", isdoublecheck);
					CHKCONFLICT("+/-F", "-h", ishelp);
					CHKCONFLICT("+/-F", "-V", isshow_version);
					if (yytext[1] == 'f')
						stripfil = FALSE;
					rcfbs |= (1L << FILCFG);
					havecmditems = TRUE;
				  }
				  return _FILOPT;
				}

<mS>G				{ if (scanpass == 1)
					isdebug = TRUE;
				  return _OPT;
				}

<mS>h				{ if (scanpass == 1) {
					ishelp = TRUE;
					CHKCONFLICT("-h", "-b", isbackup);
					CHKCONFLICT("-h", "-c", ischeck);
					CHKCONFLICT("-h", "-C", isdoublecheck);
					CHKCONFLICT("-h", "-d", isdisable);
					CHKCONFLICT("-h", "-e", isenable);
					CHKCONFLICT("-h", "-E", islist_elements);
					CHKCONFLICT("-h", "-f", isdiff);
					CHKCONFLICT("-h", "-i", isinstall);
					CHKCONFLICT("-h", "-I", isupdate_includes);
					CHKCONFLICT("-h", "-k", iskill_piktd);
					CHKCONFLICT("-h", "-K", iskill_piktc_svc);
					CHKCONFLICT("-h", "-l", islist_files);
					CHKCONFLICT("-h", "-L", islist_items);
					CHKCONFLICT("-h", "-m", iscksum);
					CHKCONFLICT("-h", "-M", isexpand_macros);
					CHKCONFLICT("-h", "-p", isinstall_piktconf);
					CHKCONFLICT("-h", "-r", isrestart_piktd);
					CHKCONFLICT("-h", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-h", "-s", isshow);
					CHKCONFLICT("-h", "-t", isdelete);
					CHKCONFLICT("-h", "-v", isverbose);
					CHKCONFLICT("-h", "-x", isexecwait);
					CHKCONFLICT("-h", "-X", isexec);
				  }
				  return _OPT;
				}

<mS,pS>H			{ BEGIN(HS);
				  if (scanpass == 1) {
					CHKCONFLICT("+/-H", "-h", ishelp);
					CHKCONFLICT("+/-H", "-V", isshow_version);
					havecmdhosts = TRUE;
				  }
				  return _HSTOPT;
				}

<mS>i				{ if (scanpass == 1) {
					isinstall = TRUE;
					CHKCONFLICT("-i", "-c", ischeck);
					CHKCONFLICT("-i", "-C", isdoublecheck);
					CHKCONFLICT("-i", "-d", isdisable);
					CHKCONFLICT("-i", "-E", islist_elements);
					CHKCONFLICT("-i", "-f", isdiff);
					CHKCONFLICT("-i", "-h", ishelp);
					CHKCONFLICT("-i", "-k", iskill_piktd);
					CHKCONFLICT("-i", "-K", iskill_piktc_svc);
					CHKCONFLICT("-i", "-l", islist_files);
					CHKCONFLICT("-i", "-L", islist_items);
					CHKCONFLICT("-i", "-m", iscksum);
					CHKCONFLICT("-i", "-M", isexpand_macros);
					CHKCONFLICT("-i", "-p", isinstall_piktconf);
					CHKCONFLICT("-i", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-i", "-s", isshow);
					CHKCONFLICT("-i", "-t", isdelete);
					CHKCONFLICT("-i", "-V", isshow_version);
					CHKCONFLICT("-i", "-x", isexecwait);
					CHKCONFLICT("-i", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>I				{ if (scanpass == 1) {
					isupdate_includes = TRUE;
					CHKCONFLICT("-I", "-h", ishelp);
					CHKCONFLICT("-I", "-k", iskill_piktd);
					CHKCONFLICT("-I", "-K", iskill_piktc_svc);
					CHKCONFLICT("-I", "-p", isinstall_piktconf);
					CHKCONFLICT("-I", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-I", "-V", isshow_version);
				  }
				  return _OPT;
				}

<mS>k				{ if (scanpass == 1) {
					iskill_piktd = TRUE;
					CHKCONFLICT("-k", "-b", isbackup);
					CHKCONFLICT("-k", "-c", ischeck);
					CHKCONFLICT("-k", "-C", isdoublecheck);
					CHKCONFLICT("-k", "-E", islist_elements);
					CHKCONFLICT("-k", "-f", isdiff);
					CHKCONFLICT("-k", "-h", ishelp);
					CHKCONFLICT("-k", "-i", isinstall);
					CHKCONFLICT("-k", "-I", isupdate_includes);
					CHKCONFLICT("-k", "-l", islist_files);
					CHKCONFLICT("-k", "-L", islist_items);
					CHKCONFLICT("-k", "-m", iscksum);
					CHKCONFLICT("-k", "-M", isexpand_macros);
					CHKCONFLICT("-k", "-p", isinstall_piktconf);
					CHKCONFLICT("-k", "-r", isrestart_piktd);
					CHKCONFLICT("-k", "-s", isshow);
					CHKCONFLICT("-k", "-V", isshow_version);
					CHKCONFLICT("-k", "-x", isexecwait);
					CHKCONFLICT("-k", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>K				{ if (scanpass == 1) {
					iskill_piktc_svc = TRUE;
					CHKCONFLICT("-K", "-b", isbackup);
					CHKCONFLICT("-K", "-c", ischeck);
					CHKCONFLICT("-K", "-C", isdoublecheck);
					CHKCONFLICT("-K", "-d", isdisable);
					CHKCONFLICT("-K", "-e", isenable);
					CHKCONFLICT("-K", "-E", islist_elements);
					CHKCONFLICT("-K", "-f", isdiff);
					CHKCONFLICT("-K", "-h", ishelp);
					CHKCONFLICT("-K", "-i", isinstall);
					CHKCONFLICT("-K", "-I", isupdate_includes);
					CHKCONFLICT("-K", "-l", islist_files);
					CHKCONFLICT("-K", "-L", islist_items);
					CHKCONFLICT("-K", "-m", iscksum);
					CHKCONFLICT("-K", "-M", isexpand_macros);
					CHKCONFLICT("-K", "-p", isinstall_piktconf);
					CHKCONFLICT("-K", "-r", isrestart_piktd);
					CHKCONFLICT("-K", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-K", "-s", isshow);
					CHKCONFLICT("-K", "-t", isdelete);
					CHKCONFLICT("-K", "-V", isshow_version);
					CHKCONFLICT("-K", "-x", isexecwait);
					CHKCONFLICT("-K", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>l				{ if (scanpass == 1) {
					islist_files = TRUE;
					CHKCONFLICT("-l", "-b", isbackup);
					CHKCONFLICT("-l", "-c", ischeck);
					CHKCONFLICT("-l", "-C", isdoublecheck);
					CHKCONFLICT("-l", "-d", isdisable);
					CHKCONFLICT("-l", "-e", isenable);
					CHKCONFLICT("-l", "-E", islist_elements);
					CHKCONFLICT("-l", "-f", isdiff);
					CHKCONFLICT("-l", "-h", ishelp);
					CHKCONFLICT("-l", "-i", isinstall);
					CHKCONFLICT("-l", "-k", iskill_piktd);
					CHKCONFLICT("-l", "-K", iskill_piktc_svc);
					CHKCONFLICT("-l", "-L", islist_items);
					CHKCONFLICT("-l", "-m", iscksum);
					CHKCONFLICT("-l", "-M", isexpand_macros);
					CHKCONFLICT("-l", "-p", isinstall_piktconf);
					CHKCONFLICT("-l", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-l", "-s", isshow);
					CHKCONFLICT("-l", "-V", isshow_version);
					CHKCONFLICT("-l", "-x", isexecwait);
					CHKCONFLICT("-l", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>L				{ if (scanpass == 1) {
					islist_items = TRUE;
					CHKCONFLICT("-L", "-b", isbackup);
					CHKCONFLICT("-L", "-c", ischeck);
					CHKCONFLICT("-L", "-C", isdoublecheck);
					CHKCONFLICT("-L", "-d", isdisable);
					CHKCONFLICT("-L", "-e", isenable);
					CHKCONFLICT("-L", "-E", islist_elements);
					CHKCONFLICT("-L", "-f", isdiff);
					CHKCONFLICT("-L", "-h", ishelp);
					CHKCONFLICT("-L", "-i", isinstall);
					CHKCONFLICT("-L", "-k", iskill_piktd);
					CHKCONFLICT("-L", "-K", iskill_piktc_svc);
					CHKCONFLICT("-L", "-l", islist_files);
					CHKCONFLICT("-L", "-m", iscksum);
					CHKCONFLICT("-L", "-M", isexpand_macros);
					CHKCONFLICT("-L", "-p", isinstall_piktconf);
					CHKCONFLICT("-L", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-L", "-s", isshow);
					CHKCONFLICT("-L", "-V", isshow_version);
					CHKCONFLICT("-L", "-x", isexecwait);
					CHKCONFLICT("-L", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>M				{ if (scanpass == 1) {
					isexpand_macros = TRUE;
					CHKCONFLICT("-M", "-b", isbackup);
					CHKCONFLICT("-M", "-c", ischeck);
					CHKCONFLICT("-M", "-C", isdoublecheck);
					CHKCONFLICT("-M", "-d", isdisable);
					CHKCONFLICT("-M", "-e", isenable);
					CHKCONFLICT("-M", "-E", islist_elements);
					CHKCONFLICT("-M", "-f", isdiff);
					CHKCONFLICT("-M", "-h", ishelp);
					CHKCONFLICT("-M", "-i", isinstall);
					CHKCONFLICT("-M", "-k", iskill_piktd);
					CHKCONFLICT("-M", "-K", iskill_piktc_svc);
					CHKCONFLICT("-M", "-l", islist_files);
					CHKCONFLICT("-M", "-L", islist_items);
					CHKCONFLICT("-M", "-m", iscksum);
					CHKCONFLICT("-M", "-p", isinstall_piktconf);
					CHKCONFLICT("-M", "-r", isrestart_piktd);
					CHKCONFLICT("-M", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-M", "-s", isshow);
					CHKCONFLICT("-M", "-t", isdelete);
					CHKCONFLICT("-M", "-V", isshow_version);
					CHKCONFLICT("-M", "-x", isexecwait);
					CHKCONFLICT("-M", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>m[1-5]			{ if (scanpass == 1) {
					iscksum = atoi(yytext+1);
					CHKCONFLICT("-m", "-b", isbackup);
					CHKCONFLICT("-m", "-c", ischeck);
					CHKCONFLICT("-m", "-C", isdoublecheck);
					CHKCONFLICT("-m", "-d", isdisable);
					CHKCONFLICT("-m", "-e", isenable);
					CHKCONFLICT("-m", "-E", islist_elements);
					CHKCONFLICT("-m", "-f", isdiff);
					CHKCONFLICT("-m", "-h", ishelp);
					CHKCONFLICT("-m", "-i", isinstall);
					CHKCONFLICT("-m", "-k", iskill_piktd);
					CHKCONFLICT("-m", "-K", iskill_piktc_svc);
					CHKCONFLICT("-m", "-l", islist_files);
					CHKCONFLICT("-m", "-L", islist_items);
					CHKCONFLICT("-m", "-M", isexpand_macros);
					CHKCONFLICT("-m", "-p", isinstall_piktconf);
					CHKCONFLICT("-m", "-r", isrestart_piktd);
					CHKCONFLICT("-m", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-m", "-s", isshow);
					CHKCONFLICT("-m", "-t", isdelete);
					CHKCONFLICT("-m", "-V", isshow_version);
					CHKCONFLICT("-m", "-x", isexecwait);
					CHKCONFLICT("-m", "-X", isexec);
				  }
				  return _OPT;
				}

<mS,pS>(O)f?			{ BEGIN(OS);
				  if (scanpass == 1) {
					CHKCONFLICT("+/-O", "-c", ischeck);
					CHKCONFLICT("+/-O", "-C", isdoublecheck);
					CHKCONFLICT("+/-O", "-h", ishelp);
					CHKCONFLICT("+/-O", "-V", isshow_version);
					if (yytext[1] == 'f')
						stripobj = FALSE;
					rcfbs |= (1L << OBJCFG);
					havecmditems = TRUE;
				  }
				  return _OBJOPT;
				}

<mS>p				{ if (scanpass == 1) {
					isinstall_piktconf = TRUE;
					CHKCONFLICT("-p", "-c", ischeck);
					CHKCONFLICT("-p", "-C", isdoublecheck);
					CHKCONFLICT("-p", "-E", islist_elements);
					CHKCONFLICT("-p", "-f", isdiff);
					CHKCONFLICT("-p", "-h", ishelp);
					CHKCONFLICT("-p", "-k", iskill_piktd);
					CHKCONFLICT("-p", "-K", iskill_piktc_svc);
					CHKCONFLICT("-p", "-l", islist_files);
					CHKCONFLICT("-p", "-L", islist_items);
					CHKCONFLICT("-p", "-m", iscksum);
					CHKCONFLICT("-p", "-M", isexpand_macros);
					CHKCONFLICT("-p", "-V", isshow_version);
					CHKCONFLICT("-p", "-x", isexecwait);
					CHKCONFLICT("-p", "-X", isexec);
				  }
				  if (scanpass == 1) {
					rcfbs |= (1L << MCRCFG);
					rcfbs |= (1L << FILCFG);
					havecmditems = TRUE;
					regname(&incmdfiles, "PIKT.conf", FALSE, FALSE);
				  }
				  return _OPT;
				}

<mS,pS>(P)f?			{ BEGIN(PS);
				  if (scanpass == 1) {
					CHKCONFLICT("+/-P", "-c", ischeck);
					CHKCONFLICT("+/-P", "-C", isdoublecheck);
					CHKCONFLICT("+/-P", "-h", ishelp);
					CHKCONFLICT("+/-P", "-V", isshow_version);
					if (yytext[1] == 'f')
						stripprg = FALSE;
					rcfbs |= (1L << PRGCFG);
					havecmditems = TRUE;
				  }
				  return _PRGOPT;
				}

<mS>r				{ if (scanpass == 1) {
					isrestart_piktd = TRUE;
					CHKCONFLICT("-r", "-c", ischeck);
					CHKCONFLICT("-r", "-C", isdoublecheck);
					CHKCONFLICT("-r", "-E", islist_elements);
					CHKCONFLICT("-r", "-f", isdiff);
					CHKCONFLICT("-r", "-h", ishelp);
					CHKCONFLICT("-r", "-k", iskill_piktd);
					CHKCONFLICT("-r", "-K", iskill_piktc_svc);
					CHKCONFLICT("-r", "-l", islist_files);
					CHKCONFLICT("-r", "-L", islist_items);
					CHKCONFLICT("-r", "-m", iscksum);
					CHKCONFLICT("-r", "-M", isexpand_macros);
					CHKCONFLICT("-r", "-p", isinstall_piktconf);
					CHKCONFLICT("-r", "-V", isshow_version);
					CHKCONFLICT("-r", "-x", isexecwait);
					CHKCONFLICT("-r", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>R				{ if (scanpass == 1) {
					isrestart_piktc_svc = TRUE;
					CHKCONFLICT("-R", "-b", isbackup);
					CHKCONFLICT("-R", "-c", ischeck);
					CHKCONFLICT("-R", "-C", isdoublecheck);
					CHKCONFLICT("-R", "-d", isdisable);
					CHKCONFLICT("-R", "-e", isenable);
					CHKCONFLICT("-R", "-E", islist_elements);
					CHKCONFLICT("-R", "-f", isdiff);
					CHKCONFLICT("-R", "-h", ishelp);
					CHKCONFLICT("-R", "-i", isinstall);
					CHKCONFLICT("-R", "-I", isupdate_includes);
					CHKCONFLICT("-R", "-K", iskill_piktc_svc);
					CHKCONFLICT("-R", "-l", islist_files);
					CHKCONFLICT("-R", "-L", islist_items);
					CHKCONFLICT("-R", "-m", iscksum);
					CHKCONFLICT("-R", "-M", isexpand_macros);
					CHKCONFLICT("-R", "-p", isinstall_piktconf);
					CHKCONFLICT("-R", "-s", isshow);
					CHKCONFLICT("-R", "-t", isdelete);
					CHKCONFLICT("-R", "-V", isshow_version);
					CHKCONFLICT("-R", "-x", isexecwait);
					CHKCONFLICT("-R", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>s				{ if (scanpass == 1) {
					isshow = TRUE;
					CHKCONFLICT("-s", "-b", isbackup);
					CHKCONFLICT("-s", "-c", ischeck);
					CHKCONFLICT("-s", "-C", isdoublecheck);
					CHKCONFLICT("-s", "-d", isdisable);
					CHKCONFLICT("-s", "-e", isenable);
					CHKCONFLICT("-s", "-E", islist_elements);
					CHKCONFLICT("-s", "-f", isdiff);
					CHKCONFLICT("-s", "-h", ishelp);
					CHKCONFLICT("-s", "-i", isinstall);
					CHKCONFLICT("-s", "-k", iskill_piktd);
					CHKCONFLICT("-s", "-K", iskill_piktc_svc);
					CHKCONFLICT("-s", "-l", islist_files);
					CHKCONFLICT("-s", "-L", islist_items);
					CHKCONFLICT("-s", "-m", iscksum);
					CHKCONFLICT("-s", "-M", isexpand_macros);
					CHKCONFLICT("-s", "-p", isinstall_piktconf);
					CHKCONFLICT("-s", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-s", "-t", isdelete);
					CHKCONFLICT("-s", "-V", isshow_version);
					CHKCONFLICT("-s", "-x", isexecwait);
					CHKCONFLICT("-s", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>t				{ if (scanpass == 1) {
					isdelete = TRUE;
					CHKCONFLICT("-t", "-c", ischeck);
					CHKCONFLICT("-t", "-C", isdoublecheck);
					CHKCONFLICT("-t", "-e", isenable);
					CHKCONFLICT("-t", "-f", isdiff);
					CHKCONFLICT("-t", "-h", ishelp);
					CHKCONFLICT("-t", "-i", isinstall);
					CHKCONFLICT("-t", "-K", iskill_piktc_svc);
					CHKCONFLICT("-t", "-m", iscksum);
					CHKCONFLICT("-t", "-M", isexpand_macros);
					CHKCONFLICT("-t", "-p", isinstall_piktconf);
					CHKCONFLICT("-t", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-t", "-s", isshow);
					CHKCONFLICT("-t", "-V", isshow_version);
					CHKCONFLICT("-t", "-x", isexecwait);
					CHKCONFLICT("-t", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>T				{ if (scanpass == 1) {
					istest = TRUE;
					/* isdebug = TRUE; */
				  }
				  return _OPT;
				}

<mS>v				{ if (scanpass == 1) {
					isverbose = TRUE;
					CHKCONFLICT("-v", "-h", ishelp);
					CHKCONFLICT("-v", "-V", isshow_version);
				  }
				  return _OPT;
				}

<mS>V				{ if (scanpass == 1) {
					isshow_version = TRUE;
					CHKCONFLICT("-V", "-b", isbackup);
					CHKCONFLICT("-V", "-c", ischeck);
					CHKCONFLICT("-V", "-C", isdoublecheck);
					CHKCONFLICT("-V", "-d", isdisable);
					CHKCONFLICT("-V", "-e", isenable);
					CHKCONFLICT("-V", "-E", islist_elements);
					CHKCONFLICT("-V", "-f", isdiff);
					CHKCONFLICT("-V", "-i", isinstall);
					CHKCONFLICT("-V", "-I", isupdate_includes);
					CHKCONFLICT("-V", "-k", iskill_piktd);
					CHKCONFLICT("-V", "-K", iskill_piktc_svc);
					CHKCONFLICT("-V", "-l", islist_files);
					CHKCONFLICT("-V", "-L", islist_items);
					CHKCONFLICT("-V", "-m", iscksum);
					CHKCONFLICT("-V", "-M", isexpand_macros);
					CHKCONFLICT("-V", "-p", isinstall_piktconf);
					CHKCONFLICT("-V", "-r", isrestart_piktd);
					CHKCONFLICT("-V", "-R", isrestart_piktc_svc);
					CHKCONFLICT("-V", "-s", isshow);
					CHKCONFLICT("-V", "-t", isdelete);
					CHKCONFLICT("-V", "-v", isverbose);
					CHKCONFLICT("-V", "-x", isexecwait);
					CHKCONFLICT("-V", "-X", isexec);
				  }
				  return _OPT;
				}

<mS>x|X				{ if (yytext[0] == 'x') {
					if (scanpass == 1) {
						CHKCONFLICT("-x", "-X", isexec);
						isexecwait = TRUE;
					}
				  } else {
					if (scanpass == 1) {
						CHKCONFLICT("-X", "-x", isexecwait);
						isexec = TRUE;
					}
				  }
				  if (scanpass == 1) {
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-b", isbackup);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-c", ischeck);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-C", isdoublecheck);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-d", isdisable);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-e", isenable);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-E", islist_elements);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-f", isdiff);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-h", ishelp);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-i", isinstall);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-k", iskill_piktd);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-K", iskill_piktc_svc);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-l", islist_files);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-L", islist_items);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-m", iscksum);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-M", isexpand_macros);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-p", isinstall_piktconf);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-r", isrestart_piktd);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-R", isrestart_piktc_svc);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-s", isshow);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-t", isdelete);
					CHKCONFLICT((yytext[0] == 'x' ? "-x" : "-X"), "-V", isshow_version);
				  }
				  return _OPT;
				}

<mS,pS>{W}			{ BEGIN(INITIAL); }

<mS,pS>.			{ if (scanpass == 1) {	/* should never get
							   to scanpass 2
							 */
					sprintf(errmsg, "illegal option: %c", yytext[0]);
					cmderror(errmsg);
					bailout = TRUE;
				  }
				  return _OPT;
				}

<AS,PS,FS,OS,HS>all		{ return _ALL; }

<INITIAL>ALL			{ for (ii=strlen(ALL)-1; ii>=0; ii--)
					unput(ALL[ii]);
				}

<DS>{DL}{nW}*			{ RETURN(_DEFINE); }

<AS>{DL}{nW}*			{ RETURN(_ALERT); }

<PS>{DL}{nW}*			{ RETURN(_PROGRAM); }

<FS>{DL}{nW}*			{ RETURN(_MYFILE); }

<OS>{DL}{nW}*			{ RETURN(_OBJECT); }

<HS>{DL}{nW}*			{ RETURN(_HOST); }

<CS>\177[^\177]*\177		{ yytext[yyleng-1] = '\0';
				  cmdlval.sval = mystrdup(yytext+1, _cmdlex_);
				  if (scanpass == 1) {
					/* we handle '\177' stripping above */
					inspectline(cmdlval.sval, &hasmacro, FALSE);
					if (hasmacro)
						rcfbs |= (1L << MCRCFG);
				  }
				  return _COMMAND;
				}

<CS>("-"|"+")[A-Z]		{ BEGIN(INITIAL);
				  yyless(0);
				}

<CS>{nW}+			{ if (scanpass == 1) {
					inspectline(yytext, &hasmacro, FALSE);
					if (hasmacro)
						rcfbs |= (1L << MCRCFG);
				  }
				  RETURN(_COMMAND);
				}

<DS,CS,AS,PS,FS,OS,HS>{W}	{ ; }

<INITIAL,DS,CS,AS,PS,FS,OS,HS>.	{ return yytext[0]; }

%%

/* cmd error handler
 */
void
cmderror (const char *message)
{

	logreturn(OUTPUT, LOG_ERR, ERRID_CMD_PARSE_ERROR, __FILE__, __LINE__, "cmdparse", FALSE, "%s: %s", message, yytext ? yytext : nullstring);
	bailout = TRUE;
	return;
}

/* feed command-line arguments to the lexer input stream
 */
signed int
cmdinput (void)
{
	register char c;

	if (argvcurr > argvlast)
		return(EOF);
	if (aoffset == 0) {	/* at beginning of new argv */
		if (strchr(*argvcurr, ' ') || strchr(*argvcurr, '\t'))
			wrapstr(argvcurr);
	}
	if ((c = (*argvcurr)[aoffset++]) != '\0')
		return(c);
	/* end of argument, unwrap() if necessary, then move on to next */
	if (**argvcurr == '\177')
		unwrapstr(argvcurr);
	argvcurr++;
	aoffset = 0;
	return(' ');
}

/* put a wrapper around the given string to signal to the lexer that this
   is a multi-part command string
 */
void
wrapstr (char **s)
{
	char *t;

	t = (char *) mymalloc((strlen(*s)+3) * sizeof(char), "wrapstr");
	sprintf(t, "%c%s%c", '\177', *s, '\177');
	*s = t;
	return;
}

/* undo the wrapping in wrapstr(), necessary because we scan the
   command line twice and must put things back as they were
 */
void
unwrapstr (char **s)
{
	char *t;

	t = mystrdup((*s)+1, "unwrapstr");
	*(t+strlen(*s)-2) = '\0';
	*s = t;
	return;
}
