/*
   ident.c -- PIKT pikct ident(ifier) routines

   $Header: /cvs/PIKT/pikt/src/piktc/ident.c,v 1.47 2002/10/03 18:58:34 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktc.h"

ident *idents[256+OBJCFG];
const char *identypes[] = { "operating system",
                            "host",
                            "hostgroup",
                            "hostalias",
                            "define",
                            "macro",
                            "alarm",
                            "alert",
                            "program",
                            "file",
                            "object"
                          };

unsigned char hash(unsigned char *, int);

/* do a lookup of an identifier in the ident list
 */
ident *
lookupident (char *identname, int cfgtype)
{
	#ifdef DEEPDEBUG
	static char fnnam[] = "lookupident";
	#endif
	ident *idp;

	foreach (idp, idents[hash(identname, cfgtype)]) {
		#ifdef DEEPDEBUG
		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "strcmping %s with %s", identname, idp->name);
		#endif
		if (! strcmp(identname,idp->name))
			return idp;
	}
	return (ident *) NULL;
}

/* register an identifier in the ident list
 */
ident *
regident (char *identname, int identype, int setdefbit)
{
	static char fnnam[] = "regident";
	ident *idp, *nidp, *aidp;
	int cfgtype;
	static unsigned long *tbs;

	#ifdef DEEPDEBUG
	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "regidenting IDENT @%s@", identname);
	#endif
	/* map the identype to its corresponding cfgtype before doing the
	   lookup:
		OPS	->	SYSCFG
		HST	->	SYSCFG
		GRP	->	SYSCFG
		ALS	->	SYSCFG
		DEF	->	DEFCFG
		MCR	->	MCRCFG
		ALM	->	ALMCFG
		ALT	->	ALTCFG
		PRG	->	PRGCFG
		FIL	->	FILCFG
		OBJ	->	OBJCFG
	 */
	cfgtype = (identype <= ALS ? SYSCFG : identype - 3);
	nidp = (ident *) mymalloc(sizeof(ident), fnnam);
	nidp->next = (ident *) NULL;
	nidp->name = identname;
	nidp->type = identype;
	nidp->alias = (char *) NULL;
	nidp->bs = (unsigned long *) NULL;
	nidp->definition = (def *) NULL;
	nidp->defcurr = (def *) NULL;
	nidp->prmnum = 0;
	nidp->called = FALSE;
	if (havekeysconf) {
		nidp->piktuid = NIL;
		nidp->piktgid = NIL;
		nidp->piktprivate_key = (char *) NULL;
	} else {
		nidp->piktuid = (int) authuidmaster;
		nidp->piktgid = (int) authgidmaster;
		nidp->piktprivate_key = pvtkeymaster;
	}
	if (lookupident(identname, cfgtype)) {
		logreturn(OUTPUT, LOG_ERR, ERRID_IDENT_ALREADY_REGISTERED, __FILE__, __LINE__, fnnam, FALSE, "%s already registered as %s", identname, identypes[identype]);
		bailout = TRUE;
		return nidp;
	}
	switch (identype) {
		case HST:	/* host */
			regname(&cfghosts, identname, FALSE, FALSE);	/* for command-line */
			*(hosts+hostnum++) = identname;	/* remember hostname */
			nidp->bs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), fnnam);
			*(nidp->bs+((hostnum-1)/(8*sizeof(unsigned long)))) = (1L << ((hostnum-1) % (8*sizeof(unsigned long))));
			tbs = hbsfull;
			hbsfull = bsunion(hbsfull, nidp->bs, hbslongs);
			free(tbs);
			break;
		case ALS:	/* host alias */
			if ((aidp = lookupident(hostcurr, SYSCFG))) {
				if (aidp->type != HST) {
					logreturn(OUTPUT, LOG_ERR, ERRID_ALIAS_IDENT_NOT_HOST, __FILE__, __LINE__, fnnam, FALSE, "%s is not a host", aidp->name);
					bailout = TRUE;
					return nidp;
				} else {
					nidp->alias = mystrdup(aidp->name, fnnam);
					nidp->bs = bsdup(aidp->bs, hbslongs);
				}
			} else {
				logreturn(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "unknown host %s", hostcurr);
				bailout = TRUE;
				return nidp;;
			}
			break;
		case OPS:	/* operating system */
			*(opsys+opsysnum++) = identname;	/* remember os name */
			nidp->bs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), fnnam);
			ghbs = nidp->bs;
			break;
		case GRP:	/* hostgroup */
			*(hstgrp+hstgrpnum++) = identname;	/* remember hostgroup name */
			nidp->bs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), fnnam);
			ghbs = nidp->bs;
			break;
		case DEF:	/* define */
			defnum++;
			nidp->bs = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), fnnam);
			*(nidp->bs+((defnum-1)/(8*sizeof(unsigned long)))) = (1L << ((defnum-1) % (8*sizeof(unsigned long))));
			if (setdefbit) {	/* set the def bit */
				tbs = gdbs;
				gdbs = bsunion(gdbs, nidp->bs, dbslongs);
				free(tbs);
			} else {		/* set the undef bit */
				tbs = gubs;
				gubs = bsunion(gubs, nidp->bs, dbslongs);
				free(tbs);
			}
			/* dbsfull currently serves no purpose
			tbs = dbsfull;
			dbsfull = bsunion(dbsfull, nidp->bs, dbslongs);
			free(tbs);
			 */
			break;
		case MCR:	/* macro */
		case ALM:	/* alarm */
			/* nidp->definition = (def *) NULL; */
			break;
		default:	/* ALT/PRG/FIL/OBJ--alert/program/file/object */
			break;
	}
	for (idp=idents[hash(identname, cfgtype)]; idp && idp->next; idp=idp->next) ;
	if (! idp)
		idents[hash(identname, cfgtype)] = nidp;
	else
		idp->next = nidp;
	return nidp;
}

/* garbage collect all processed idents DEF and higher (i.e., everything
   except what's in systems.cfg)
 */
void
cleanup (void)
{
	#ifdef DEEPDEBUG
	static char fnnam[] = "cleanup";
	#endif
	register int i;
	static ident *idp, *tidp;
	static def *dfp, *tdfp;

	for (i=0; i<=255+OBJCFG; i++) {
		if (! idents[i])
			continue;
		idp = tidp = idents[i];
		while (idp && (idp->type < DEF)) {
			tidp = idp;
			idp = idp->next;
		}
		if (tidp != idp)
			tidp->next = (ident *) NULL;
		if (idp == idents[i])
			idents[i] = (ident *) NULL;
		while (idp) {
			tidp = idp;
			idp = idp->next;
			if (tidp->name) {
				#ifdef DEEPDEBUG
				if (isdebug)
					logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "freeing IDENT %s", tidp->name);
				#endif
				free(tidp->name);
			}
			if (tidp->alias) {
				#ifdef DEEPDEBUG
				if (isdebug)
					logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "freeing IDENT alias %s", tidp->alias);
				#endif
				free(tidp->alias);
			}
			if (tidp->bs) {
				free(tidp->bs);
			}
			dfp = tidp->definition;
			while (dfp) {
				tdfp = dfp;
				dfp = dfp->next;
				if (tdfp->component) {
					#ifdef DEEPDEBUG
					if (isdebug)
						logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "freeing COMP %s", tdfp->component);
					#endif
					free(tdfp->component);
				}
				free(tdfp);
			}
			free(tidp);
		}
	}
	return;
}

/* sum the ident's first and last chars and add the cfgtype (to allow for
   multiple idents instances in the hash table, but just one per config file
   type)
 */
unsigned char
hash (unsigned char *idnt, int cfgtype)
{
	static char fnnam[] = "hash";
	unsigned char c1, c2, c;
	register int i;

	i = 0;
	c1 = *idnt;
	while (*(idnt+i++)) ;
	c2 = *(idnt+i-2);
	c = c1 + c2 + cfgtype;
	if (c > 255+OBJCFG) {
		logreturn(OUTPUT, LOG_ERR, ERRID_HASH_INVALID, __FILE__, __LINE__, fnnam, FALSE, "for ident %s, invalid hash val %d", idnt, c);
		bailout = TRUE;
		return 0;	/* ERROR stand-in */
	}
	return c;
}

/* make the built-in, predefined pikt macros
 */
void
makepiktmacros (char *host)
{
	static char fnnam[] = "makepiktmacros";
	ident *idp;
	char uid[64], gid[64];			/* should be plenty */
	char private_key[2*MAXKEYBYTES+3];
	register int i;

	idp = regident(mystrdup("pikthostname", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(hostcurr, fnnam));

	/* we have to do these again for each hostcurr,
	   since macros are eliminated in each cleanup()
	 */

	idp = regident(mystrdup("piktdir", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(homdir, fnnam));

	idp = regident(mystrdup("piktmaster", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(master, fnnam));

	idp = regident(mystrdup("pikthosts", fnnam), MCR, FALSE);
	for (i=0; i<hostnum; i++)
		regcomponent(idp, mystrdup(*(hosts+i), fnnam));

	idp = regident(mystrdup("piktopsystems", fnnam), MCR, FALSE);
	for (i=0; i<opsysnum; i++)
		regcomponent(idp, mystrdup(*(opsys+i), fnnam));

	if (! (idp = lookupident(host, SYSCFG)))
		logexit(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupident() failure for %s", host);
	sprintf(uid, "%d", (idp->piktuid == NIL) ? (int) authuidmaster : idp->piktuid);
	sprintf(gid, "%d", (idp->piktgid == NIL) ? (int) authgidmaster : idp->piktgid);
	sprintf(private_key, "\"%s\"", (! idp->piktprivate_key) ? pvtkeymaster : idp->piktprivate_key);
	idp = regident(mystrdup("piktuid", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(uid, fnnam));
	idp = regident(mystrdup("piktgid", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(gid, fnnam));
	idp = regident(mystrdup("piktprivate_key", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(private_key, fnnam));

	idp = regident(mystrdup("piktnever", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(PIKTNEVER, fnnam));

	idp = regident(mystrdup("piktnullchar", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(piktnullchar, fnnam));

	idp = regident(mystrdup("_", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(piktnullchar, fnnam));

	idp = regident(mystrdup("piktspacechar", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(piktspacechar, fnnam));

	idp = regident(mystrdup("pikttabchar", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(pikttabchar, fnnam));

	idp = regident(mystrdup("piktnewlinechar", fnnam), MCR, FALSE);
	regcomponent(idp, mystrdup(piktnewlinechar, fnnam));

	return;
}

/* set path to the default "/bin/nice" (which is accurate for most
   systems); if nice has been defined as a macro, if that macrodef is
   a filepath, and if that file exists, then set path to that filepath
   instead; combine with prio to make nicecmd
 */
char *
getnicecmd (char *prio)
{
	static char fnnam[] = "getnicecmd";
	ident *idp;
	struct stat fbuf;
	const char *path;
	char cmd[MAXCMD+1];
	char *r;

	path = "/bin/nice";	/* default path */
	if ((idp = lookupident("nice", MCRCFG))) {
		if (lstat(idp->definition->component, &fbuf) == 0)
			path = idp->definition->component;
	}
	sprintf(cmd, "%s -%s", path, prio);
	r = mystrdup(cmd, fnnam);
	return r;
}

/* for every os and hostgroup explicitly defined in systems.cfg, make an
   implicit "non" version
 */
/*
void
makenongroups (char **identnames, int identcnt)
{
	char *identname;
	int i;

	for (i=0; i<identcnt; i++) {
		identname = *(identnames+i);
		makenongroup(identname);
	}
	return;
}
*/

/* for the given os or hostgroup, make an implicit "non" version, consisting
   of the host complement; e.g., if "linux" is a list of all linux systems,
   "nonlinux" is a list of all systems not linux; if "nonlinux" is explicitly
   defined in systems.cfg, that definition overrides any implicit definition
   that would otherwise be made here
 */
void
makenongroup (char *identname)
{
	static char fnnam[] = "makenongroup";
	char *nonname;
	int idlen;
	ident *idp, *nidp;
	unsigned long *cbs;

	if (! (idp = lookupident(identname, SYSCFG))) {
		logreturn(OUTPUT, LOG_ERR, ERRID_IDENT_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "ident %s not found", identname);
		bailout = TRUE;
		return;
	}
	if ((idlen = strlen(identname)) > MAXIDENT-3) {
		logreturn(OUTPUT, LOG_ERR, ERRID_IDENT_MAX_LENGTH_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "ident %s too long to make non group", identname);
		bailout = TRUE;
		return;
	}
	nonname = (char *) mymalloc((size_t)(idlen+3+1) * sizeof(char), fnnam);
	sprintf(nonname, "non%s", identname);
	if (lookupident(nonname, SYSCFG)) {
		logreturn(PIKTLOG, LOG_WARNING, ERRID_IDENT_ALREADY_REGISTERED, __FILE__, __LINE__, fnnam, FALSE, "ident %s already defined", nonname);
		free(nonname);
		return;
	}
	if (! (nidp = regident(nonname, GRP, FALSE))) {
		logreturn(OUTPUT, LOG_ERR, ERRID_IDENT_REGISTRATION_ERROR, __FILE__, __LINE__, fnnam, FALSE, "regident() failure for ident %s", nonname);
		bailout = TRUE;
		free(nonname);
		return;
	}
	free(nidp->bs);
	cbs = bscomplement(idp->bs, hbslongs);
	nidp->bs = bsintersection(hbsfull, cbs, hbslongs);
	free(cbs);
	return;
}
