/*
   piktc.c -- PIKT master control program, and config file compiler

   $Header: /cvs/PIKT/pikt/src/piktc/piktc.c,v 1.120 2002/10/24 14:38:44 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   --------------------------------------------------------------------------

   Usage:  piktc <-bcCdeEfGhiIkKlLm#MprRstTvVxX>
                 [+|-D          <define(s)>]
                 [+C            <command(s)>]
                 [+M            <macro(s)>]
                 [+|-A|S[f] all|<alert(s)/script(s)>]
                 [+|-P[f]   all|<program(s)>]
                 [+|-F[f]   all|<file(s)>]
                 [+|-O[f]   all|<object(s)>]
                  +|-H      all|<host(s)>
                 [ALL]

           -b                         backup alert/script/program/file/object
                                        file(s)
           -c                         syntax check all config files
           -C                         syntax doublecheck all config files
           -d                         disable alert(s)
           -e                         enable alert(s)
           -E                         list all config elements
           -f                         diff alert/script/program/file/object
                                        file(s)
           -G                         run in debug mode
           -h                         show program help
           -i                         install alert/script/program/file/object
                                        file(s)
           -I                         update #include/#verbatim files
           -k                         kill alert daemon (piktd)
           -K                         kill service daemon (pikt_svc)
           -l                         list alert/script/program/file/object
                                        file(s)
           -L                         list alert/script/program/file/object/ or
                                        host/os/group/alias command-line item(s)
           -m#                        checksum alert/script/program/file/object
                                        file(s), where # is checksum level 1-5
           -M                         expand macros
           -p                         install PIKT.conf
           -r                         (re)start alert daemon (piktd)
           -R                         (re)start service daemon (piktc_svc)
           -s                         show alert(s) status
           -t                         delete alert/script/program/file/object/
                                        history/log file(s)
           -T                         run in test mode
           -v                         run in verbose mode
           -V                         show version info
           -x                         execute alert(s)/script(s)
           -X                         execute alert(s)/script(s) with no wait
           +|-D          <define(s)>  define/undefine define(s)
           +C            <command(s)> include command string(s)
           +M            <macro(s)>   include macro string(s)
           +|-A|S[f] all|<alert(s)>   include/exclude (fix) alert(s)/script(s)
           +|-P[f]   all|<program(s)> include/exclude (fix) program file(s)
           +|-F[f]   all|<file(s)>    include/exclude (fix) other file(s)
           +|-O[f]   all|<object(s)>  include/exclude (fix) object file(s)
           +|-H      all|<host(s)>    include/exclude host(s)/os(es)/group(s)/
                                        alias(es)
           ALL                        +A all +S all +P all +F all +O all +H all

 */

#include "../lib/common.h"
#include "piktc.h"

#include <rpc/rpc.h>
#include "rpc.h"

const char piktc_version[] = "PIKT piktc version " VERSION;

AUTHUNIX_CREATE_UID_T authuidmaster = 0;
AUTHUNIX_CREATE_GID_T authgidmaster = 0;
char *pvtkeymaster = (char *) NULL;

/* command-line options */
int isbackup = FALSE;
int ischeck = FALSE;
int isdebug = FALSE;
int isdelete = FALSE;
int iscksum = FALSE;
int isdiff = FALSE;
int isdoublecheck = FALSE;
int isdisable = FALSE;
int isenable = FALSE;
int isexec = FALSE;
int isexecwait = FALSE;
int isexpand_macros = FALSE;
int ishelp = FALSE;
int isinstall = FALSE;
int isupdate_includes = FALSE;
int iskill_piktd = FALSE;
int iskill_piktc_svc = FALSE;
int islist_elements = FALSE;
int islist_files = FALSE;
int islist_items = FALSE;
int isinstall_piktconf = FALSE;
int isrestart_piktd = FALSE;
int isrestart_piktc_svc = FALSE;
int isshow = FALSE;
int isshow_version = FALSE;
int istest = FALSE;
int isverbose = FALSE;

int bailout = FALSE;
int bypass = FALSE;

int scanpass;

char **argvcurr;	/* current command-line argument */
char **argvlast;	/* last command-line argument */

const char nullstring[] = { '\0' };

const signed char err[]    = { ERROR, '\0' };
const signed char nil[]    = { NIL, '\0' };
const char errstr[] = "<<ERR>>";
const char nilstr[] = "<<NIL>>";

/* const signed char dquote[] = { DQTPROXY, '\0' }; */
/* const signed char squote[] = { SQTPROXY, '\0' }; */
/* const char space[]  = " "; */

const char piktnullchar[]    =       { '\0' };
const char piktspacechar[]   = { ' ',  '\0' };
const char pikttabchar[]     = { '\t', '\0' };
const char piktnewlinechar[] = { '\n', '\0' };

unsigned loffset = 0;

int cfgerr = FALSE;
int stzerr = FALSE;

unsigned long cfbs = 0L;	/* initially, no config files detected yet */
unsigned long rcfbs = 1L;	/* only systems.cfg is required */
unsigned long mcfbs = 0L;	/* initially, no config files are considered
				   missing
				 */

size_t hbslongs;
size_t dbslongs;

int hostnum = 0;
int opsysnum = 0;
int hstgrpnum = 0;
int defnum = 0;

char **hosts;
char *hostcurr;

char **opsys;
char **hstgrp;

unsigned long *hbscurr = (unsigned long *) NULL;

/* cmdcommands/cmdalerts/cmdprograms/cmdfiles/cmdobjects/cmdhosts are the
   commands/alerts/programs/files/objectsets/hosts specified in the command line

   cfgalerts/cfgprograms/cfgfiles/cfgobjects are the
   alerts/programs/files/objectsets included in the tagged config files

   commands/alerts/programs/files/objects are the
   commands/alerts/programs/files/objectsets being written (or executed)
   for the current host in the current program run

   since we bailout if any alert/program/file/objectsset (or host) specified
   in the command line is bogus (meaning it's not found in the config files),
   cmdalerts/cmdprograms/cmdfiles/cmdobjects is a subset of
   cfgalerts/cfgprograms/cfgfiles/cfgobjects

   commands/alerts/programs/files/objects is identical to
   cmdcommands/cmdalerts/cmdprograms/cmdfiles/cmdobjects if comand-line
   commands/alerts/programs/files/objects are specified, else
   cmdcommands/cmdalerts/cmdprograms/cmdfiles/cmdobjects is empty and
   commands/alerts/programs/files/objects are the things actually written
   (or executed)
 */

nam *commands,               *incmdcommands,                 *cmdcommands,
    *alerts,   *cfgalerts,   *incmdalerts,   *excmdalerts,   *cmdalerts,
    *programs, *cfgprograms, *incmdprograms, *excmdprograms, *cmdprograms,
    *files,    *cfgfiles,    *incmdfiles,    *excmdfiles,    *cmdfiles,
    *objects,  *cfgobjects,  *incmdobjects,  *excmdobjects,  *cmdobjects,
               *cfghosts,    *incmdhosts,    *excmdhosts,    *cmdhosts;

int commandsall, alertsall, programsall,
    filesall, objectsall, hostsall;

int havecmditems = FALSE;
int havecmdhosts = FALSE;

FILE	*stripin,	*stripout,
	*tagin,		*tagout,
	*sysin,		*sysout,
	*cfgin,		*cfgout;

FILE *keysin;
int havekeysconf = FALSE;

char *inproc = (char *) NULL;
char *infile = (char *) NULL;

char *inctmpfile = (char *) NULL;
FILE *inctmpout = (FILE *) NULL;
int incdevnull = FALSE;

int fd[2];
pid_t pid;
int lineno;

/* the order of the following .cfg files is vitally important! */
const char *cfgfilenames[] = { "systems.cfg",
                               "defines.cfg",
                               "macros.cfg",
                               "alarms.cfg",
                               "alerts.cfg",
                               "programs.cfg",
                               "files.cfg",
                               "objects.cfg"
                             };

char *cfgfileprev = (char *) NULL;
char *cfgfilecurr = (char *) NULL;

int cfgcurr;
int cfglast;

cfgln *configs[OBJCFG+1];	/* assumes OBJCFG is last cfg file */

/* the order of the following is vitally important! */
const char *cfgbof[] = {        "/\n",	/* SYSCFG */
                               "//\n",	/* DEFCFG */
                              "///\n",	/* MCRCFG */
                             "////\n",	/* ALMCFG */
                            "/////\n",	/* ALTCFG */
                           "//////\n",	/* PRGCFG */
                          "///////\n",	/* FILCFG */
                         "////////\n"	/* OBJCFG */
                       };

char *cfgline;

cfgln *clplast = (cfgln *) NULL;

int indented = FALSE;
int verbatim = FALSE;

unsigned bof;
unsigned bon;
unsigned bol;

int timings;

char *mins[MAXTIMINGS+1], *hrs[MAXTIMINGS+1],
     *doms[MAXTIMINGS+1], *mons[MAXTIMINGS+1],
     *dows[MAXTIMINGS+1];
char *drift[MAXTIMINGS+1];
char *priority;
char *nicecmd, *mailcmd, *lpcmd;
char *execcmd;

int stripalm, stripprg, stripfil, stripobj;

int hasmacro;
int domacros;
int dostrip;

int incnum;
int inctype;

char piktcprog[13];			/* PIKTCPROG string */
char piktcvers[5];			/* PIKTCVERS string; PIKTCVERS <= 99 */

/* the ( after for and switch below allow for _FORLP ('for (') & _SWITCHLP
   ('switch (') in cfg.l; parentheses following 'if' and others are handled
   differently and should NOT appear in the following keywords[] strings
 */
const char keywords_sections[] = " init begin rule end ";
const char keywords_statements[] = " status level task input filter seps separators keys dat data delete do if elif elsif elseif else endif fi while endwhile repeat until for ( foreach endfor endforeach switch ( endswitch endsw breakswitch breaksw case default break cont continue again redo next skip last pause quit die output set unset exec ";
const char keywords_statements_block_begin[] = " default if elif elsif elseif else while repeat for ( foreach switch ( case ";
const char keywords_statements_block_end[] = " default elif elsif elseif else endif fi endwhile until endfor endforeach endswitch endsw case ";

unsigned long services = ALL_SERVICES;	/* all services enabled by default */

unsigned int fork_delay = 1;

void preprocess(int, int, int);
void processcmd(int, char **);
void process(char *, unsigned long*, int, char **);
void mkcmds(char *);
void dodoublecheck(void);
void getkeys(void);
int missingcfgfiles(int);

int
main (int argc, char *argv[])
{
	static char fnnam[] = "main";
	static char pidstr[8];

	if (argc < 2) {		/* called simply as 'piktc' (no args) */
		piktc_usage();
		exit(0);
	}

	init(argc, argv, (void (*)(void))piktc_usage, piktc_version);
	if (bailout)
		exit(exstat);

	if (    ! isfile(cpcmd, TRUE)
	     || ! isfile(diffcmd, TRUE)
	     || ! isfile(rmcmd, TRUE)
	     || ! isfile(egrepcmd, TRUE)
	     || ! isfile(rpcinfocmd, TRUE)
	   ) {
		/* error(s) logged in PIKTLOG */
		exit(exstat);
	}

	if (isfile(keysconf, FALSE))
		havekeysconf = TRUE;

	authuidmaster = authuid;
	authgidmaster = authgid;
	pvtkeymaster = pvtkey;

	/* initializations for rpcinfo probes */
	sprintf(piktcprog, " %ld ", (long)PIKTCPROG);
	sprintf(piktcvers, " %d ", PIKTCVERS);

	commands = incmdcommands = cmdcommands =
	alerts = cfgalerts = incmdalerts = excmdalerts = cmdalerts =
	programs = cfgprograms = incmdprograms = excmdprograms = cmdprograms =
	files = cfgfiles = incmdfiles = excmdfiles = cmdfiles =
	objects = cfgobjects = incmdobjects = excmdobjects = cmdobjects =
	          cfghosts = incmdhosts = excmdhosts = cmdhosts = (nam *)NULL;
	stripalm = stripprg = stripfil = stripobj = TRUE;

	argvcurr = argv+1;	/* skip the program name */
	argvlast = argv+argc-1;

	scanpass = 1;
	cmdparse();		/* further parse the command line */
	if (bailout) {
		piktc_usage();
		exit(exstat);
	}

	/* apply the lock, to prevent other concurrent piktc sessions,
	   but only if we are not doing a 'piktc -L +H ... -H ...', or
	   under certain circumstances, 'piktc -x' or 'piktc -X', which
	   don't require any writing to the staging directory, thus present
	   no problems when run concurrently with other piktc actions;
	   note: we must ensure that cmdparse() is called before this,
	   giving the opportunity for islist_items, havecmditems, &
	   havecmdhosts to be set; also init(), to determine PIKT.conf
	   settings
	 */
	if (! (    (islist_items && ! havecmditems && havecmdhosts && ! isupdate_includes)
	        || (isexecwait && isupdate_includes)
	        || (isexpand_macros)
	      )
	   ) {
		/* check for blocking lockfile */
		if (islock(piktclock)) {
			logexit(OUTPUT, LOG_ERR, ERRID_MULTIPLE_PIKTC, __FILE__, __LINE__, fnnam, FALSE, "%s found!  Another piktc session may be in progress.  Try again later (else rm the %s lockfile).", piktclock, piktclock);
		}
		/* check for running piktc */
		if (procid("piktc", PROCID) == ERROR) {
			logexit(OUTPUT, LOG_ERR, ERRID_MULTIPLE_PIKTC, __FILE__, __LINE__, fnnam, FALSE, "Another piktc session may be in progress.  Try again later (else kill the running piktc and/or rm the %s lockfile, if any).", piktclock);
		}
		/* apply the lock */
		pid = getpid();
		sprintf(pidstr, "%u", (unsigned)pid);
		if (mklock(piktclock, pidstr, TRUE, TRUE) == ERROR)
			logexit(PIKTLOG, LOG_ERR, ERRID_FILE_CREATE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "mklock() error for %s", piktclock);
	}

	/* when using '#include|#verbatim [piktc -xI ...]', we now need to
	   turn off the '-I' updates, else we will wind up in an endless
	   update loop; with 'piktc -xI', the '-I' is effectively just a way
	   of bypassing the lock file mechanism, therefore
	 */
	if (isexecwait && isupdate_includes)
		isupdate_includes = FALSE;

	/* preprocess the config files; the order is vitally important! */

	/* always preprocess systems.cfg */
	if (! bailout) preprocess(SYSCFG, TRUE, 1);
	if (! bailout) preprocess(SYSCFG, TRUE, 2);
	/*
	if (! bailout) makenongroups(opsys, opsysnum);
	*/
	if (! bailout) preprocess(SYSCFG, TRUE, 3);

	/* process keys.conf, if present */
	if (! bailout) getkeys();

	/* we can't tell from the command line if preprocessing the
	   defines.cfg file is necessary, since defines might be set and
	   referred to later on in other config files, so we always
	   preprocess defines.cfg (possibly just a faked "dummy" file)
	 */
	if (! bailout) preprocess(DEFCFG, TRUE, 1);
	if (! bailout) preprocess(DEFCFG, TRUE, 2);

	/* if we have 'piktc -x|X +C "<commands without macros>" +H ...',
	   we don't need to process any more config files
	 */
	if ((isexec || isexecwait) && ! ((rcfbs & (1L << MCRCFG)) || (rcfbs & (1L << ALTCFG)))) {
		;	/* no further preprocessing required */
	} else {
		if (! bailout) preprocess(MCRCFG, TRUE, 2);
		if (! bailout && missingcfgfiles(MCRCFG)) {
			bailout = TRUE;
		/* if we have 'piktc -x|X +C "<commands with macros>" +H ...',
		   or if we have 'piktc -M +C "<commands|string>" +H ...',
		   we don't need to process any more config files
		 */
		} else if ((isexec || isexecwait || isexpand_macros) && (rcfbs < (1L << ALMCFG))) {
			;	/* no further preprocessing required */
		} else {
			/* if we have 'piktc -c|C ...' */
			/* or if we have 'piktc ... +A|S ...' */
			if (ischeck || isdoublecheck || islist_elements || rcfbs & (1L << ALMCFG))
				if (! bailout) preprocess(ALMCFG, stripalm, 2);
			/* or if we have 'piktc ... +A|S ...' */
			if (ischeck || isdoublecheck || islist_elements || rcfbs & (1L << ALTCFG))
				if (! bailout) preprocess(ALTCFG, TRUE, 2);
			/* or if we have 'piktc ... +P ...' */
			if (ischeck || isdoublecheck || islist_elements || rcfbs & (1L << PRGCFG))
				if (! bailout) preprocess(PRGCFG, stripprg, 2);
			/* or if we have 'piktc ... +F ...' */
			if (ischeck || isdoublecheck || islist_elements || rcfbs & (1L << FILCFG))
				if (! bailout) preprocess(FILCFG, stripfil, 2);
			/* or if we have 'piktc ... +O ...' */
			if (ischeck || isdoublecheck || islist_elements || rcfbs & (1L << OBJCFG))
				if (! bailout) preprocess(OBJCFG, stripobj, 2);
			if (! bailout && missingcfgfiles(OBJCFG))
				bailout = TRUE;
		}
	}

	/* preprocess command-line items */
	if (! bailout) preprocesscmd();
	
	/* execute the command(s) */
	if (! bailout) processcmd(argc, argv);

	/*
	if (bailout) piktc_usage();
	 */

	/* only remove the lock if not doing a 'piktc -L' */
	if (! (islist_items && ! havecmditems && havecmdhosts)) {
		/* remove the lock */
		if (rmlocks(TRUE, TRUE) == ERROR)
			logexit(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "rmlocks() error");
	}

	exit(exstat);
}

/* output program usage to screen
 */
void
piktc_usage (void)
{
	printf("\nUsage:  piktc <-bcCdeEfGhiIkKlLm#MprRstTvVxX>\n\t      [+|-D          <define(s)>]\n\t      [+C            <command(s)>]\n\t      [+M            <macro(s)>]\n\t      [+|-A|S[f] all|<alert(s)/script(s)>]\n\t      [+|-P[f]   all|<program(s)>]\n\t      [+|-F[f]   all|<file(s)>]\n\t      [+|-O[f]   all|<object(s)>]\n\t       +|-H      all|<host(s)>\n\t      [ALL]\n\n");
	printf("\t-b\t\t\t   backup alert/script/program/file/object\n\t\t\t\t     file(s)\n");
	printf("\t-c\t\t\t   syntax check all config files\n");
	printf("\t-C\t\t\t   syntax doublecheck all config files\n");
	printf("\t-d\t\t\t   disable alert(s)\n");
	printf("\t-e\t\t\t   enable alert(s)\n");
	printf("\t-E\t\t\t   list all config elements\n");
	printf("\t-f\t\t\t   diff alert/script/program/file/object\n\t\t\t\t     file(s)\n");
	printf("\t-G\t\t\t   run in debug mode\n");
	printf("\t-h\t\t\t   show program help\n");
	printf("\t-i\t\t\t   install alert/script/program/file/object\n\t\t\t\t     file(s)\n");
	printf("\t-I\t\t\t   update #include/#verbatim files\n");
	printf("\t-k\t\t\t   kill alert daemon (piktd)\n");
	printf("\t-K\t\t\t   kill service daemon (piktc_svc)\n");
	printf("\t-l\t\t\t   list alert/script/program/file/object\n\t\t\t\t     file(s)\n");
	printf("\t-L\t\t\t   list alert/script/program/file/object or\n\t\t\t\t     host/os/group/alias command-line item(s)\n");
	printf("\t-m#\t\t\t   checksum alert/script/program/file/object\n\t\t\t\t     file(s), where # is checksum level 1-5\n");
	printf("\t-M\t\t\t   expand macros\n");
	printf("\t-p\t\t\t   install PIKT.conf\n");
	printf("\t-r\t\t\t   (re)start alert daemon (piktd)\n");
	printf("\t-R\t\t\t   (re)start service daemon (piktc_svc)\n");
	printf("\t-s\t\t\t   show alert(s) status\n");
	printf("\t-t\t\t\t   delete alert/script/program/file/object\n\t\t\t\t     /history/log file(s)\n");
	printf("\t-T\t\t\t   run in test mode\n");
	printf("\t-v\t\t\t   run in verbose mode\n");
	printf("\t-V\t\t\t   show version info\n");
	printf("\t-x\t\t\t   execute alert(s)/script(s)\n");
	printf("\t-X\t\t\t   execute alert(s)/script(s) with no wait\n");
	printf("\t+|-D          <define(s)>  define/undefine define(s)\n");
	printf("\t+C            <command(s)> include command string(s)\n");
	printf("\t+M            <macro(s)>   include macro string(s)\n");
	printf("\t+|-A|S[f] all|<alert(s)>   include/exclude (fix) alert(s)\n");
	printf("\t+|-P[f]   all|<program(s)> include/exclude (fix) program file(s)\n");
	printf("\t+|-F[f]   all|<file(s)>    include/exclude (fix) other file(s)\n");
	printf("\t+|-O[f]   all|<object(s)>  include/exclude (fix) object file(s)\n");
	printf("\t+|-H      all|<host(s)>    include/exclude host(s)/os(es)/group(s)/\n\t\t\t\t     alias(es)\n");
	printf("\tALL\t\t\t   +A all +S all +P all +F all +O all +H all\n\n");
	return;
}

/* for each host in the host-bit set, process the piktc command: execute
   commands/alarms/scripts, enable/disable alerts, install/delete files,
   list files, kill/restart alert daemon, show alert status, check config
   file validity
   note: "NIL" is a dummy extension for signalling unambiguously that a file
   is a program from programs.cfg, not from files.cfg
 */
void
processcmd (int argc, char **argv)
{
	static char fnnam[] = "processcmd";
	ident *idp;
	nam *hstp;
	char cmd[MAXCMD+1];
	static unsigned long *hbs;
	static unsigned long *tbs;
	unsigned long hbsl;
	register int i, j;
	struct stat fbuf;

	/* for each host in the command-line set, turn on its corresponding
	   bit in the host bitset
	 */
	hbs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), fnnam);
	foreach (hstp, cmdhosts) {
		idp = lookupident(hstp->name, SYSCFG);
		if (idp) {
			tbs = hbs;
			hbs = bsunion(hbs, idp->bs, hbslongs);
			free(tbs);
		} else	/* validatecmditems() should have caught this */ {
			logreturn(OUTPUT, LOG_WARNING, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "unknown host or group:  %s", hstp->name);
		}
	}
	/* foreach unsigned long in the host bitset */
	for (i=0; i<hbslongs; i++) {
		hbsl = *(hbs+i);
		/* foreach bit in the current unsigned long */
		for (j=0; j<8*sizeof(unsigned long); j++) {
			#ifdef DEEPDEBUG
			if (isdebug)
				logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "i is %d, j is %d, 1L<<j is %x, hbsl is %x", i, j, 1L<<j, hbsl);
			#endif
			/* if the bit for this host is on */
			if (hbsl & (1L<<j)) {
				hostcurr = *(hosts+((i*sizeof(unsigned long)*8)+j));
				FREE(hbscurr);
				hbscurr = bsdup(getbs(hostcurr, HST, FALSE), hbslongs);
				cfgerr = FALSE;
				stzerr = FALSE;

				if (islist_items && ! havecmditems) {
					printf("%s\n", hostcurr);
					fflush(stdout);
					continue;
				}

				if (isverbose) {
					printf("\n%s%sing %s...\n", isdoublecheck ? "double" : nullstring, (ischeck || isdoublecheck) ? "check" : "process", hostcurr);
					fflush(stdout);
				}

				/*
				if (isverbose && islist_elements) {
					printf("listing elements...\n");
					fflush(stdout);
				}
				 */

				/* if not doing a syntax check, determine if
				   remote piktc_svc is running; if not, say
				   so, and bypass the current host
				 */
				if (! (ischeck || isdoublecheck || islist_elements || islist_items || isexpand_macros || isinstall_piktconf)) {
					if (chkpiktc_svc(hostcurr) != 0) {
						if (isverbose) {
							logreturn(OUTPUT, LOG_WARNING, ERRID_PIKTC_SVC_NOT_RUNNING, __FILE__, __LINE__, fnnam, FALSE, "remote piktc_svc not running (or unspecified rpc failure)");
						}
						continue;
					}
				}

				if ((idp = lookupident(hostcurr, SYSCFG))) {
					authuid = (AUTHUNIX_CREATE_UID_T) (idp->piktuid == NIL) ? authuidmaster : idp->piktuid;
					authgid = (AUTHUNIX_CREATE_GID_T) (idp->piktgid == NIL) ? authgidmaster : idp->piktgid;
					pvtkey = (! idp->piktprivate_key) ? pvtkeymaster : idp->piktprivate_key;
					pvtkeylen = pvtkey ? strlen(pvtkey) : 0;
				} else {
					logexit(OUTPUT, LOG_ERR, ERRID_IDENT_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupident() failure for %s", hostcurr);
				}

				/* disable remote alerts */
				if ((/* isinstall || */ isdisable || isenable || isdelete) && cmdalerts)
					disable_alerts(hostcurr, alertsall ? cfgalerts : cmdalerts, isverbose && ! isenable /* && ! isinstall */);

				/* delete remote alert files */
				if ((/* isinstall || */ isdelete) && cmdalerts)
					delete_files(hostcurr, alertsall ? cfgalerts : cmdalerts, ALTDIR, ".alt", isbackup, isverbose && ! isinstall);

				/* delete remote history/log files */
				if (isdelete && cmdalerts) {
					delete_files(hostcurr, alertsall ? cfgalerts : cmdalerts, HSTDIR, ".hst", isbackup, isverbose && ! isinstall);
					delete_files(hostcurr, alertsall ? cfgalerts : cmdalerts, LOGDIR, ".log", isbackup, isverbose && ! isinstall);
				}

				/* delete remote object files */
				if ((/* isinstall || */ isdelete) && cmdobjects)
					delete_files(hostcurr, objectsall ? cfgobjects : cmdobjects, OBJDIR, ".obj", isbackup, isverbose && ! isinstall);

				/* install/enable/execute/check/
				   doublecheck/diff cfg/dat files */
				if (isinstall || isenable || isexec || isexecwait || isexpand_macros || ischeck || isdoublecheck || iscksum || isdiff || islist_elements || islist_items || isinstall_piktconf || (islist_files && (cmdprograms || cmdfiles)) || (isdelete && (cmdprograms || cmdfiles))) {

					/* 'piktc -x' and 'piktc -X' should not
					   touch the diffing or staging dirs
					   in any way
					 */
					if (! (isexec || isexecwait || isexpand_macros)) {

						/* clear out diffing dir */
						if (isdiff) {
							sprintf(cmd, "%s -rf %s 2>/dev/null", rmcmd, difdir);
							system(cmd);
							/* wait for difdir rm */
							while (stat(difdir, &fbuf) == 0)
								sleep(1);
							if (mkdir(difdir, (mode_t)(S_IRWXU|S_IRGRP|S_IWGRP)) == -1)	/* 750 */
								logexit(OUTPUT, LOG_ERR, ERRID_MKDIR_ERROR, __FILE__, __LINE__, fnnam, TRUE, "mkdir() error for %s", difdir);
						}

						/* clear out staging dir */
						sprintf(cmd, "%s -rf %s 2>/dev/null", rmcmd, stgdir);
						system(cmd);
						/* wait for stgdir rm */
						while (stat(stgdir, &fbuf) == 0)
							sleep(1);
						if (mkdir(stgdir, (mode_t)(S_IRWXU|S_IRGRP|S_IWGRP)) == -1)	/* 750 */
							logexit(OUTPUT, LOG_ERR, ERRID_MKDIR_ERROR, __FILE__, __LINE__, fnnam, TRUE, "mkdir() error for %s", stgdir);

					}

					/* process config files for current host */
					process(hostcurr, hbscurr, argc, argv);

					/* garbage collect; we can't do this
					   yet in the install_piktconf case,
					   because in install_piktconf(), we
					   invoke piktx, which requires macros
					   defined for the current host
					 */
					if (! isinstall_piktconf)
						cleanup();

					/* if cfgparse() error, just bypass
					   current host, no longer bail out
					 */
					if (bypass) {
						bypass = FALSE;
						retire(&commands);
						retire(&alerts);
						retire(&programs);
						retire(&files);
						retire(&objects);
						continue;
					}

					if (isdoublecheck) {
						dodoublecheck();
					}

					if (cfgerr || ischeck || isdoublecheck || islist_elements)
						continue;

				}

				/* note: it is necessary to delete program
				   and other files after invoking process(),
				   for we need to know the programs and other
				   files list for the current machine;
				   otherwise, for programs and other files
				   installed outside the default PIKT program
				   directory and inside a directory NFS
				   cross-mounted to more than one machine,
				   we risk blowing them away by passing
				   cfgprograms or cfgfiles to delete_files()
				   (in the manner we do above) when processing
				   a later machine
				 */

				/* delete remote program files */
				if ((/* isinstall || */ isdelete) && cmdprograms)
					delete_files(hostcurr, programs, PRGDIR, "NIL", isbackup, isverbose && ! isinstall);

				/* delete remote other files */
				if ((/* isinstall || */ isdelete) && cmdfiles)
					delete_files(hostcurr, files, nullstring, nullstring, isbackup, isverbose && ! isinstall);

				/* note: an alert/program/file/object may be
				   specified in the command line but not be
				   written to the staging directory, because
				   it doesn't apply to a particular host,
				   which is why we can't fetch/install/enable
				   cmdalerts/cmdprograms/cmdfiles/cmdobjects,
				   which might include
				   alerts/programs/files/objects not processed
				 */

				/* install remote alert files */
				if (isinstall && cmdalerts)
					install_files(hostcurr, alerts, ALTDIR, ".alt", MODE_FIL_DEFAULT, UID_FIL_DEFAULT, GID_FIL_DEFAULT, isbackup, isverbose);

				/* install remote program files */
				if (isinstall && cmdprograms)
					install_files(hostcurr, programs, PRGDIR, "NIL", MODE_PRG_DEFAULT, UID_FIL_DEFAULT, GID_FIL_DEFAULT, isbackup, isverbose);

				/* install remote other files */
				if (isinstall && cmdfiles)
					install_files(hostcurr, files, nullstring, nullstring, MODE_FIL_DEFAULT, UID_FIL_DEFAULT, GID_FIL_DEFAULT, isbackup, isverbose);

				/* install remote object files */
				if (isinstall && cmdobjects)
					install_files(hostcurr, objects, OBJDIR, ".obj", MODE_FIL_DEFAULT, UID_FIL_DEFAULT, GID_FIL_DEFAULT, isbackup, isverbose);

				/* cksum remote alert files */
				if (iscksum && cmdalerts)
					cksum_files(hostcurr, alerts, ALTDIR, ".alt", iscksum, isverbose);

				/* cksum remote program files */
				if (iscksum && cmdprograms)
					cksum_files(hostcurr, programs, PRGDIR, "NIL", iscksum, isverbose);

				/* cksum remote other files */
				if (iscksum && cmdfiles)
					cksum_files(hostcurr, files, nullstring, nullstring, iscksum, isverbose);

				/* cksum remote object files */
				if (iscksum && cmdobjects)
					cksum_files(hostcurr, objects, OBJDIR, ".obj", iscksum, isverbose);

				/* fetch remote alert files */
				if (isdiff && cmdalerts)
					fetch_files(hostcurr, alerts, ALTDIR, ".alt", isverbose);

				/* fetch remote program files */
				if (isdiff && cmdprograms)
					fetch_files(hostcurr, programs, PRGDIR, "NIL", isverbose);

				/* fetch remote other files */
				if (isdiff && cmdfiles)
					fetch_files(hostcurr, files, nullstring, nullstring, isverbose);

				/* fetch remote object files */
				if (isdiff && cmdobjects)
					fetch_files(hostcurr, objects, OBJDIR, ".obj", isverbose);

				/* diff the diffing & staging dirs */
				if (isdiff) {
					if (isverbose) {
						printf("diffing file(s)...\n");
						fflush(stdout);
					}
					sprintf(cmd, "%s -r %s %s | %s -iv \"common subdir\"", diffcmd, stgdir, difdir, egrepcmd);
					system(cmd);
				}

				/* enable remote alerts */
				if ((/* isinstall || */ isenable) && cmdalerts)
					enable_alerts(hostcurr, alertsall ? alerts : alerts, isverbose);

				/* execute remote commands */
				if ((isexec || isexecwait) && cmdcommands)
					execute_commands(hostcurr, commandsall ? commands : commands, isverbose);

				/* execute remote alerts */
				if ((isexec || isexecwait) && cmdalerts)
					execute_alerts(hostcurr, alertsall ? alerts : alerts, isverbose);

				/* expand macros */
				if (isexpand_macros && cmdcommands)
					expand_macros(hostcurr, commandsall ? commands : commands, isverbose);

				/* list remote alert files */
				if (islist_files && cmdalerts)
					list_files(hostcurr, alertsall ? cfgalerts: cmdalerts, ALTDIR, ".alt", isverbose);

				/* list remote program files */
				if (islist_files && cmdprograms)
					list_files(hostcurr, programsall ? cfgprograms : cmdprograms, PRGDIR, "NIL", isverbose);

				/* list remote other files */
				if (islist_files && cmdfiles)
					list_files(hostcurr, filesall ? cfgfiles : cmdfiles, nullstring, nullstring, isverbose);

				/* list remote object files */
				if (islist_files && cmdobjects)
					list_files(hostcurr, objectsall ? cfgobjects : cmdobjects, OBJDIR, ".obj", isverbose);

				/* list command-line alert items */
				if (islist_items && alerts)
					list_items(hostcurr, alerts, ALTDIR, nullstring, isverbose);

				/* list command-line program items */
				if (islist_items && programs)
					list_items(hostcurr, programs, PRGDIR, "NIL", isverbose);

				/* list command-line other items */
				if (islist_items && files)
					list_items(hostcurr, files, nullstring, nullstring, isverbose);

				/* list command-line object items */
				if (islist_items && objects)
					list_items(hostcurr, objects, OBJDIR, nullstring, isverbose);

				/* kill remote alert daemon (piktd) */
				if (iskill_piktd || isrestart_piktd)
					kill_daemon(hostcurr, KILL_PIKTD_SERVICE, isverbose && ! isrestart_piktd);

				/* kill remote service daemon (piktc_svc) */
				if (iskill_piktc_svc)
					kill_daemon(hostcurr, KILL_PIKTC_SVC_SERVICE, isverbose && ! isrestart_piktc_svc);

				/* install remote PIKT.conf */
				if (isinstall_piktconf) {
					install_piktconf(hostcurr, MODE_FIL_DEFAULT, UID_FIL_DEFAULT, GID_FIL_DEFAULT, isbackup, isverbose);
					/* garbage collect */
					cleanup();
				}

				/* (re)start remote alert daemon (piktd) */
				if (isrestart_piktd) {
					/* allow time for daemon to die from
					   just-concluded kill
					 */
					sleep(1+fork_delay);
					restart_daemon(hostcurr, RESTART_PIKTD_SERVICE, isverbose);
				}

				/* (re)start remote service daemon (piktc_svc) */
				if (isrestart_piktc_svc) {
					/* allow time for daemon to die from
					   just-concluded kill
					 */
					sleep(1+fork_delay);
					restart_daemon(hostcurr, RESTART_PIKTC_SVC_SERVICE, isverbose);
				}

				/* show remote alert status */
				if (isshow && cmdalerts)
					show_alerts_stata(hostcurr, alertsall ? cfgalerts : cmdalerts, isverbose);

				retire(&commands);
				retire(&alerts);
				retire(&programs);
				retire(&files);
				retire(&objects);
			}
		}
	}
	/* bypass the request deregistration when doing 'piktc -L +H' or
	   'piktc -M', especially when piktc is invoked on a slave system,
	   otherwise an "unauthenticated register_request()" error message
	   will be logged in the slave's piktc_svc.log because the slave
	   is of course not the master as defined in its PIKT.conf
	 */
	if (! (    (islist_items && ! havecmditems && havecmdhosts && ! isupdate_includes)
	        || (isexpand_macros)
	      )
	   ) {
		/* deregister; this is needed here especially in the
		   isexec case, but should be redundant for everything else
		 */
		register_request(nullstring, nullstring, nullstring, FALSE);
	}
	free(hbs);
	return;
}

/* open the config file, filter out C++ style comments in strip, and
   pass the output to the sys parser for systems.cfg, and to the
   cfg parser for all other .cfg files; after preprocessing, we no
   longer need refer to config files on disk, as all information is
   stored in internal data structures
 */
void
preprocess (int cfgtype, int strip, int scnpas)
{
	static char fnnam[] = "preprocess";
	struct stat fbuf;
	unsigned long *hbs;

	/* check for existence of cfg file; in certain cases, bailout if it
	   doesn't exist
	 */
	cfgfileprev = (char *) NULL;
	cfgfilecurr = infile = (char *) mymalloc((strlen(cfgdir)+strlen(cfgfilenames[cfgtype])+2) * sizeof(char), fnnam);
	sprintf(infile, "%s/%s", cfgdir, cfgfilenames[cfgtype]);
	inproc = (char *) NULL;
	inctmpfile = (char *) NULL;
	inctmpout = (FILE *) NULL;
	if (stat(infile, &fbuf) != 0) {
		switch (cfgtype) {
			/* systems.cfg will always be mandatory
			 */
			case SYSCFG:
				logexit(OUTPUT, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", infile);
				break;
			case DEFCFG:
				/* even if no actual defines.cfg, fake one so
				   as to avoid a possible reduce/reduce
				   error in def.y that might result if all
				   config files, except systems.cfg, are
				   optional
				 */
				if (scnpas == 2) {
					cfgcurr = cfgtype;
					hbs = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), fnnam);
					fill(hbs, hbslongs);
					tag(1, cfgfilecurr, hbs, hbslongs, FALSE, FALSE, "dummy FALSE\n");
					FREE(hbs);
					cfbs |= (1L << cfgtype);
					lineno = 1;
					logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", infile);
				}
				return;
				break;
			case MCRCFG:
			case ALMCFG:
			case ALTCFG:
			case PRGCFG:
			case FILCFG:
			case OBJCFG:
				logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() error for %s", infile);
				return;
				break;
		}
	} else {
		if (cfgtype > SYSCFG)
			cfbs |= (1L << cfgtype);	/* add this cfg file
							   to config bitset
						 	 */
		else
			cfbs = 1L;
	}
	cfgcurr = cfgtype;
	dostrip = strip;
	scanpass = scnpas;
	if (scanpass == 2) {
		switch (cfgtype) {
			case SYSCFG:
				hbslongs = (int)((hostnum-1)/(8*sizeof(long))) + 1;
				hosts = (char **) mymalloc(hostnum*sizeof(char *), fnnam);
				opsys = (char **) mymalloc(opsysnum*sizeof(char *), fnnam);
				hstgrp = (char **) mymalloc(hstgrpnum*sizeof(char *), fnnam);
				hostnum = 0;
				opsysnum = 0;
				hstgrpnum = 0;
				hbsfull = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), fnnam);
				hbsempty = (unsigned long *) mycalloc(hbslongs, sizeof(unsigned long), fnnam);
				break;
			case DEFCFG:
				dbslongs = (int)((defnum-1)/(8*sizeof(long))) + 1;
				gdbs = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), fnnam);
				gubs = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), fnnam);
				/* currently serve no purpose
				dbsfull = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), fnnam);
				dbsempty = (unsigned long *) mycalloc(dbslongs, sizeof(unsigned long), fnnam);
				 */
				break;
			default:
				break;
		}
	}
	domacros = TRUE;
	incnum = 0;
	inctype = INCFILE;
	lineno = 1;
	indented = verbatim = FALSE;	/* reset before each new config
					   file type
					 */
	if (dostrip) {
		if (pipe(fd) < 0)
			logexit(OUTPUT, LOG_ERR, ERRID_PIPE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "pipe() error");
		/* open the ipc channels */
		ipcopen(incnum, __FILE__, __LINE__);
		switch (pid = fork()) {
			case -1:	/* error */
				logexit(OUTPUT, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fork() error");
				break;
			/* pipe from child to parent */
			case 0:		/* child */
				close(fd[0]);
				stripin = fopen(infile, "r");
				stripout = fdopen(fd[1], "w");
				wait_parent(incnum, __FILE__, __LINE__);
				tell_parent(incnum, __FILE__, __LINE__);
				striplex();
				fclose(stripin);
				fclose(stripout);
				tell_parent(incnum, __FILE__, __LINE__);
				wait_parent(incnum, __FILE__, __LINE__);
				close(fd[1]);
				exit(0);
				break;
			default:	/* parent */
				close(fd[1]);
				/* the following sleep() is probably not needed
				   but helps prevent hangs on slower, overloaded
				   systems
				sleep(fork_delay);
				 */
				switch (cfgtype) {
					case SYSCFG:
						sysin = fdopen(fd[0], "r");
						tell_child(incnum, __FILE__, __LINE__);
						wait_child(incnum, __FILE__, __LINE__);
						sysparse();
						fclose(sysin);
						break;
					default:
						tagin = fdopen(fd[0], "r");
						tell_child(incnum, __FILE__, __LINE__);
						wait_child(incnum, __FILE__, __LINE__);
						tagparse();
						fclose(tagin);
						break;
				}
				wait_child(incnum, __FILE__, __LINE__);
				tell_child(incnum, __FILE__, __LINE__);
				close(fd[0]);
				break;
		}
		/* close the ipc channels */
		ipcclose(incnum, __FILE__, __LINE__);
	} else {
		/* for now, only ALMCFG/PRGCFG/FILCFG/OBJCFG are potentially
		   not stripped, SYSCFG is always stripped, so switch
		   not really necessary here
		 */
		switch (cfgtype) {
			case SYSCFG:
				sysin = fopen(infile, "r");
				sysparse();
				fclose(sysin);
				break;
			default:
				tagin = fopen(infile, "r");
				tagparse();
				fclose(tagin);
				break;
		}
	}
	return;
}

/* foreach macros, alarms, alerts, programs, files & objects config file,
   process the config lines; foreach config line, if its tag matches for
   the current host, then input it to the cfg parser
 */
void
process (char *host, unsigned long *hbs, int argc, char **argv)
{
	static char fnnam[] = "process";
	int i;
	int lastcfg;
	struct stat fbuf;

	#ifdef DEEPDEBUG
	if (isdebug)
		dumpcfgs(hbs);
	#endif

	/* determine the last config file to be processed (the first is
	   always sytems.cfg)
	 */
	lastcfg = 0;	/* systemsc.fg (SYSCFG = 0) always required */
	for (i=1; i<=OBJCFG; i++) {
		if (cfbs & (1L << i))
			lastcfg = i;
	}

	/* only call defparse() if there is a defines.cfg file
	 */
	cfgfileprev = (char *) NULL;
	cfgfilecurr = infile = (char *) mymalloc((strlen(cfgdir)+strlen(cfgfilenames[DEFCFG])+2) * sizeof(char), fnnam);
	sprintf(infile, "%s/%s", cfgdir, cfgfilenames[DEFCFG]);
	inproc = (char *) NULL;
	inctmpfile = (char *) NULL;
	inctmpout = (FILE *) NULL;
	if (stat(infile, &fbuf) == 0) {
		empty(gdbs, dbslongs);
		empty(gubs, dbslongs);
		defnum = 0;
		loffset = 0;
		cfgcurr = DEFCFG;
		cfglast = DEFCFG;
		bof = TRUE;
		defparse();
		if (bailout) {	/* set in deferror() */
			if (rmlocks(TRUE, TRUE) == ERROR)
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "rmlocks() error");
			exit(exstat);
		}
		/* now go back and apply any command-line def overrides */
		argvcurr = argv+1;	/* skip the program name */
		argvlast = argv+argc-1;
		scanpass = 2;
		cmdparse();
		if (bailout) {		/* the first scan should have
					   caught any errors
					 */
			/*
			piktc_usage();
			 */
			exit(exstat);
		}
		loffset = 0;
		cfgcurr = MCRCFG;
		cfglast = lastcfg;
		bof = TRUE;
		defparse();
		#ifdef DEEPDEBUG
		/* only dump a second time after the lines are marked
		   for doproc
		 */
		if (isdebug)
			dumpcfgs(hbs);
		#endif
		if (bailout) {	/* set in deferror() */
			if (rmlocks(TRUE, TRUE) == ERROR)
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_REMOVE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "rmlocks() error");
			exit(exstat);
		}
	}

	loffset = 0;
	cfgcurr = MCRCFG;
	cfglast = lastcfg;
	cfgfileprev = (char *) NULL;
	cfgfilecurr = (char *) cfgfilenames[MCRCFG];
	bof = TRUE;
	bon = FALSE;
	bol = FALSE;
	makepiktmacros(host);
	cfgparse();
	mkcmds(host);
	return;
}

/* for every cmdcommands (if any), macro expand the command string, and
   assign to the commands nam list
 */
void
mkcmds (char *host)
{
	static char fnnam[] = "mkcmds";
	nam *nmp1, *nmp2, *nnmp;
	char *expcmd;

	foreach (nmp1, cmdcommands) {
		expcmd = expand(nmp1->name);
		nnmp = (nam *) mymalloc(sizeof(nam), fnnam);
		nnmp->next = (nam *) NULL;
		nnmp->name = mystrdup(nmp1->name, fnnam);
		nnmp->aux = expcmd;
		nnmp->dir = (char *) NULL;
		nnmp->file = (char *) NULL;
		nnmp->mode = (char *) NULL;
		nnmp->filuid = (char *) NULL;
		nnmp->filgid = (char *) NULL;
		nnmp->state = IGNORE;
		/* add to commands list */
		if (commands) {
			for (nmp2=commands; nmp2->next; nmp2=nmp2->next) ;
			nmp2->next = nnmp;
		} else
			commands = nnmp;
	}
	return;
}

/* do a second syntax check on each alert in stgdir, by running them through
   a pikt -C check (which does a more thorough syntax check than piktc);
   if problems are discovered in the pikt doublecheck, it might be nice to
   save each specimen for later inspection, but we'll reserve implementation
   of that for another time!
 */
void
dodoublecheck (void)
{
	static char fnnam[] = "dodoublecheck";
	nam *nmp;
	static char *alt = (char *) NULL;
	char cmd[MAXCMD+1];
	struct stat fbuf;

	foreach (nmp, alerts) {
		if (nmp->name) {
			FREE(alt);
			alt = (char *) mymalloc((strlen(stgdir)+strlen(nmp->name)+6) * sizeof(char), fnnam);
			sprintf(alt, "%s/%s.alt", stgdir, nmp->name);
			if (lstat(alt, &fbuf) == 0) {	/* alert file found */
				sprintf(cmd, "%s/pikt -C +A %s", bindir, nmp->name);
				system(cmd);
			}
		}
	}
	FREEVOID(alt);
	return;
}

/* check to see if piktc_svc on host is up; return 0 if so, ERROR if down;
   in the case of 'timed out' replies, we reattempt a number of times
 */
int
chkpiktc_svc (char *host)
{
	static char fnnam[] = "chkpiktc_svc";
	char cmd[MAXCMD+1];
	char line[MAXLINE+1];
	FILE *rpcinfo;
	char *cp;
	int try;

	/* for some mysterious reason, in older versions of AIX only, a
	   successful popen() call below randomly returns empty output; this
	   is an aftereffect of the IPC piping in preprocess(); this problem
	   appears to have been solved in later versions of AIX; the bug is
	   confirmed up until at least AIX 4.1.5, and is confirmed solved in
	   AIX 4.3.3; we can't say between those two reference points; what
	   follows is a workaround for this bug (where we just automatically
	   report successful contact here, but at the risk of crashing piktc
	   later on when and if it attempts to do RPC with a downed system,
	   or an up system but with no running piktc_svc); if you encounter
	   the difficulties described in README.AIX, activate the workaround
	   by uncommenting the '#ifdef AIX ... #endif' following
	 */
	/*
	#ifdef AIX
	return 0;
	#endif
	 */

	/* note: before, we had been using 'rpcinfo -t ...' successfully,
	   until irix balked (timed out with no response); so we switched
	   to 'rpcinfo -u ...', which seems to make everyone happy
	 */

	/* in earlier verions of AIX, e.g., 4.1.5 and before, a command like

	   # rpcinfo -u `hostname` 806422610

	   would return output like

	   program 806422610 version 1
	           ready and waiting

	   piping to the tr command was our kludgy way around this anomaly;
	   later versions of AIX appear not to have this glitch, so we now
	   remove the tr kludge by default

	sprintf(cmd, "%s -u %s %lu | %s '\n' ' ' 2>&1", rpcinfocmd, host, (unsigned long) RPCPRGNUM, trcmd);
	 */

	sprintf(cmd, "%s -u %s %lu 2>&1", rpcinfocmd, host, (unsigned long) RPCPRGNUM);
	try = 0;
	line[0] = '\0';
	while (++try <= 1+rpcinfo_retries) {
		if ((rpcinfo = popen(cmd, "r")) == NULL) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, FALSE, "popen() error");
			return ERROR;
		}
		if (isverboselog || isdebug)
			logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "probing rpcinfo on %s", host);
		while (fgets(line, MAXLINE+1, rpcinfo) != NULL) {
			if ((cp = strchr(line, '\n')) != NULL)
				*cp = '\0';
			if (isverboselog || isdebug)
				logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s", line);

			if (strstr(line, piktcprog) && strstr(line, piktcvers)) {
				pclose(rpcinfo);
				return 0;
			}
		}
		pclose(rpcinfo);
	}
	if ((cp = strchr(line, '\n')) != NULL)
		*cp = '\0';
	logreturn(PIKTLOG, LOG_WARNING, ERRID_RPCINFO_ERROR, __FILE__, __LINE__, fnnam, FALSE, "for remote host %s: %s", host, line[0] ? line : "unknown rpcinfo failure");
	return ERROR;
}

/* process the keys.conf file, if any
 */
void
getkeys (void)
{
	static char fnnam[] = "getkeys";

	if ((keysin = fopen(keysconf, "r")) != NULL) {
		lineno = 1;
		keysparse();
		if (fclose(keysin)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", keysconf);
			/* bailout = TRUE; */
		}
	}
	free(keysconf);
	return;
}

/* check to see if any required config files are missing
 */
int
missingcfgfiles (int lastcfg)
{
	static char fnnam[] = "missingcfgfiles";
	register int i;
	unsigned long tcfbs;

	tcfbs = 0L;
	for (i=0; i<=lastcfg; i++)
		tcfbs |= (1L << i);
	tcfbs &= rcfbs;
	if ((mcfbs = (cfbs & tcfbs) ^ tcfbs)) {
		for (i=0; i<=lastcfg; i++) {
			if (! (mcfbs ^ (1L << i)))
				logreturn(PIKTLOG, LOG_ERR, ERRID_CFG_FILE_MISSING, __FILE__, __LINE__, fnnam, FALSE, "missing %s file", cfgfilenames[i]);
		}
		return TRUE;
	}
	return FALSE;
}
