/*
   rpcclnt.c -- PIKT rpc client (piktc) routines

   $Header: /cvs/PIKT/pikt/src/piktc/rpcclnt.c,v 1.89 2002/10/03 18:58:34 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"

#include <rpc/rpc.h>

#if defined(HAVE_NETCONFIG_H)
#include <netconfig.h>
#endif

#include <rpc/xdr.h>
#include <rpc/clnt.h>
#include "rpc.h"

#include "piktc.h"

#define RETURN_RES		switch (*res) { \
					case 0: \
						return TRUE; \
						break; \
					case UNAUTHORIZED: \
						return UNAUTHORIZED; \
						break; \
					default: \
						return FALSE; \
						break; \
				}

#define RETURN_RES_RESULT	switch (res->result) { \
					case 0: \
						return TRUE; \
						break; \
					case UNAUTHORIZED: \
						return UNAUTHORIZED; \
						break; \
					default: \
						return FALSE; \
						break; \
				}

#define RETURN_KRES		switch (kres) { \
					case UNAUTHORIZED: \
						return UNAUTHORIZED; \
						break; \
					case NONE: \
						return NONE; \
						break; \
					default: \
						return FALSE; \
						break; \
				}

CLIENT *clnt;

/* when we recompile this file to generate svc_rpc_clnt.o for linking into
   piktc_svc, we only need the call_back() function, so we put the
   '#ifndef PIKTC_SVC' wrapper around the code we don't need for piktc_svc,
   and the '#else' around the code we do need
 */

#ifndef PIKTC_SVC	/* piktc */
itm command;
itm alert;
fil _file;
#endif

req reqserv;		/* piktc & piktc_svc */

const char *transports[] = { "udp", "tcp" };

char *key = (char *) NULL;
unsigned int keylen = 0;
char *keyenc = (char *) NULL;
char *keysum = (char *) NULL;
blf_ctx bc;

#ifndef PIKTC_SVC	/* piktc */
int enable_alert(char *, char *, char *, int);
int disable_alert(char *, char *, const char *, int);
int execute_command(char *, char *, char *, int);
int execute_alert(char *, char *, char *, int);
extern int install_file(char *, char *, char *, const char *, char *, char *, char *, int, int);
int fetch_file(char *, char *, char *, const char *, int);
int cksum_file(char *, char *, char *, const char *, int, int);
int delete_file(char *, char *, char *, const char *, int, int);
int list_file(char *, char *, char *, const char *, int, int);
int show_alert_status(char *, char *, const char *, int);
int piktc_svc_pid(char *, int);
int get_key(char **, unsigned int *, CLIENT *);
int deckey (char *, unsigned int, char **, unsigned int *);
#endif

#ifndef PIKTC_SVC	/* piktc */

/* enable all alerts in the supplied list on the given host
 */
void
enable_alerts (char *host, nam *alts, int inform)
{
	nam *np;
	int res;

	register_request(host, "edit_piktd_conf", "alerts(s) not enabled", inform);
	if (inform) {
		printf("enabling alert(s)...\n");
		fflush(stdout);
	}
	foreach (np, alts) {
		if (strstr(np->aux, PIKTNEVER))	/* bypass non-alarm scripts */
			continue;
		if (np->state == ERROR) {
			if (inform) {
				printf("%s not enabled, unknown alert\n", np->name);
				fflush(stdout);
			}
			continue;
		}
		if (! (res = list_file(host, np->name, ALTDIR, ".alt", IGNORE, inform)) && (res != IGNORE)) {
			if (inform) {
				printf("%s not enabled, %s.alt not found\n", np->name, np->name);
				fflush(stdout);
			}
			continue;
		}
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("alert(s) not enabled\n");
				fflush(stdout);
			}
			break;
		}
		res = enable_alert(host, np->name, np->aux, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("alert(s) not enabled\n");
				fflush(stdout);
			}
			break;
		}
		if (inform) {
			printf("%s %senabled\n", np->name, res == 1 ? nullstring : "not ");
			fflush(stdout);
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* disable all alerts in the supplied list on the given host
 */
void
disable_alerts (char *host, nam *alts, int inform)
{
	nam *np;
	int res;

	register_request(host, "edit_piktd_conf", "alerts(s) not disabled", inform);
	if (inform) {
		printf("disabling alert(s)...\n");
		fflush(stdout);
	}
	foreach (np, alts) {
		if (np->state == ERROR) {
			if (inform) {
				printf("%s not enabled, unknown alert\n", np->name);
				fflush(stdout);
			}
			continue;
		}
		res = disable_alert(host, np->name, nullstring, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("alert(s) not disabled\n");
				fflush(stdout);
			}
			break;
		}
		if (inform) {
			printf("%s %sdisabled\n", np->name, res == 1 ? nullstring : "not ");
			fflush(stdout);
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* execute all alerts in the supplied list on the given host
 */
void
execute_commands (char *host, nam *cmds, int inform)
{
	nam *np;
	int res;

	register_request(host, "execute_command", "command(s) not executed", inform);
	if (inform) {
		printf("executing command(s)...\n");
		fflush(stdout);
	}
	foreach (np, cmds) {
		if (np->state == ERROR) {
			if (inform) {
				printf("%s not executed, unknown command\n", np->name);
				fflush(stdout);
			}
			continue;
		}
		res = execute_command(host, np->name, np->aux, inform);
		if (((res == NONE) && isexecwait) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("command(s) not executed\n");
				fflush(stdout);
			}
			break;
		}
		if (isexecwait && inform) {
			printf("\"%s\" %sexecuted\n", np->name, res == 1 ? nullstring : "not ");
			fflush(stdout);
		}
	}
	/* if we deregister the execute_command request immediately, a client
	   piktc_svc sending an asynchronous callback might find nothing
	   registered; in the isexec case, we do the deregistration in
	   piktc.c at the conclusion of all host processing
	 */
	if (isexec) {
		;	/* wait until all clients are processed */
	} else /* isexecwait, or default */ {
		register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	}
	return;
}

/* execute all alerts in the supplied list on the given host
 */
void
execute_alerts (char *host, nam *alts, int inform)
{
	nam *np;
	int res;

	register_request(host, "execute_alert", "alerts(s)/script(s) not executed", inform);
	if (inform) {
		printf("executing alert(s)/script(s)...\n");
		fflush(stdout);
	}
	foreach (np, alts) {
		if (np->state == ERROR) {
			if (inform) {
				printf("%s not executed, unknown alert/script\n", np->name);
				fflush(stdout);
			}
			continue;
		}
		res = execute_alert(host, np->name, np->aux, inform);
		if (((res == NONE) && isexecwait) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("alert(s)/script(s) not executed\n");
				fflush(stdout);
			}
			break;
		}
		if (isexecwait && inform) {
			printf("%s %sexecuted\n", np->name, res == 1 ? nullstring : "not ");
			fflush(stdout);
		}
	}
	/* if we deregister the execute_alert request immediately, a client
	   piktc_svc sending an asynchronous callback might find nothing
	   registered; in the isexec case, we do the deregistration in
	   piktc.c at the conclusion of all host processing
	 */
	if (isexec) {
		;	/* wait until all clients are processed */
	} else /* isexecwait, or default */ {
		register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	}
	return;
}

/* expand macros in all supplied command/strings for the given host
 */
void
expand_macros (char *host, nam *cmds, int inform)
{
	nam *np;

	if (inform) {
		printf("expanding command(s)/macros(s)...\n");
		fflush(stdout);
	}
	foreach (np, cmds) {
		fprintf(stdout, "%s\n", np->aux);
		fflush(stdout);
		if (inform) {
			printf("\"%s\" expanded\n", np->name);
			fflush(stdout);
		}
	}
	return;
}

/* install all files in the supplied list on the given host
 */
void
install_files (char *host, nam *fils, char *dir, const char *ext, char *mode, char *filuid, char *filgid, int backup, int inform)
{
	nam *np;
	int res;

	register_request(host, "install_file", "file(s) not installed", inform);
	if (inform) {
		printf("installing file(s)...\n");
		fflush(stdout);
	}
	foreach (np, fils) {
		res = install_file(host, np->name, dir, ext, mode, filuid, filgid, backup, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("file(s) not installed\n");
				fflush(stdout);
			}
			break;
		} else if (res == IGNORE) {
			continue;
		} else {
			if (inform) {
				printf("%s%s %sinstalled\n", np->name, strcmp(ext, "NIL") ? ext : nullstring, res == 1 ? nullstring : "not ");
				fflush(stdout);
			}
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* fetch all files in the supplied list from the given host
 */
void
fetch_files (char *host, nam *fils, char *dir, const char *ext, int inform)
{
	nam *np;
	int res;

	/* before we switched to using execute_command() to do a "cat <file>",
	   we used to register_request() a "fetch_file" here; now we pass
	   "fetch_file" as the command.name to execute_command()
	 */
	register_request(host, "execute_command", "file(s) not fetched", inform);
	if (inform) {
		printf("fetching file(s)...\n");
		fflush(stdout);
	}
	foreach (np, fils) {
		res = fetch_file(host, np->name, dir, ext, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("file(s) not fetched\n");
				fflush(stdout);
			}
			break;
		} else if (res == IGNORE) {
			continue;
		} else {
			if (inform) {
				printf("%s%s %sfetched\n", np->name, strcmp(ext, "NIL") ? ext : nullstring, res == 1 ? nullstring : "not ");
				fflush(stdout);
			}
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* checksum-and-compare all files in the supplied list on the given host
 */
void
cksum_files (char *host, nam *fils, char *dir, const char *ext, int level, int inform)
{
	nam *np;
	int res;

	register_request(host, "cksum_file", "file(s) not checksummed", inform);
	if (inform) {
		printf("checksumming file(s)...\n");
		fflush(stdout);
	}
	foreach (np, fils) {
		res = cksum_file(host, np->name, dir, ext, level, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("file(s) not checksummed\n");
				fflush(stdout);
			}
			break;
		} else if (res == IGNORE) {
			continue;
		} else {
			if (inform) {
				printf("%s%s %schecksummed\n", np->name, strcmp(ext, "NIL") ? ext : nullstring, res == 1 ? nullstring : "not ");
				fflush(stdout);
			}
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* delete all files in the supplied list on the given host
 */
void
delete_files (char *host, nam *fils, char *dir, const char *ext, int backup, int inform)
{
	nam *np;
	int res;

	register_request(host, "delete_file", "file(s) not deleted", inform);
	if (inform) {
		printf("deleting file(s)...\n");
		fflush(stdout);
	}
	foreach (np, fils) {
		res = delete_file(host, np->name, dir, ext, backup, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("file(s) not deleted\n");
				fflush(stdout);
			}
			break;
		} else if (res == IGNORE) {
			continue;
		} else {
			if (inform) {
				printf("%s%s %sdeleted\n", np->name, strcmp(ext, "NIL") ? ext : nullstring, res == 1 ? nullstring : "not ");
				fflush(stdout);
			}
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* list all files in the supplied list on the given host
 */
void
list_files (char *host, nam *fils, char *dir, const char *ext, int inform)
{
	nam *np;
	int res;

	register_request(host, "list_file", "file(s) not listed", inform);
	if (inform) {
		printf("listing file(s)...\n");
		fflush(stdout);
	}
	foreach (np, fils) {
		res = list_file(host, np->name, dir, ext, NIL, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("file(s) not listed\n");
				fflush(stdout);
			}
			break;
		} else if (res == IGNORE) {
			continue;
		} else {
			/* if (inform) */ {
				printf("%s%s %sfound\n", np->name, strcmp(ext, "NIL") ? ext : nullstring, res == 1 ? nullstring : "not ");
				fflush(stdout);
			}
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* list all command-line items in the supplied list for the given host;
   note that since this is just for processing and reporting command-line
   items and nothing else, there is no need for a list_item() function to
   contact the slave
 */
void
list_items (char *host, nam *fils, char *dir, const char *ext, int inform)
{
	nam *np;

	if (inform) {
		printf("listing item(s)...\n");
		fflush(stdout);
	}
	foreach (np, fils) {
		printf("%s%s\n", np->name, strcmp(ext, "NIL") ? ext : nullstring);
		fflush(stdout);
	}
	return;
}

/* install PIKT.conf on the given host; restart the remote piktc_svc after
   the installation
 */
void
install_piktconf (char *host, char *mode, char *filuid, char *filgid, int backup, int inform)
{
	static char fnnam[] = "install_piktconf";
	ident *etcdiridp;
	char *hostetcdir;
	char cmd[MAXCMD+1];
	int pid1, pid2;

	/* we can't call piktc_svc_pid(), or indeed any other function that
	   calls the remote piktc_svc, because any changes for the current
	   host in keys.conf will create mismatch(es) with keys still defined,
	   for the moment, in the remote PIKT.conf
	pid1 = piktc_svc_pid(host, FALSE);
	   so we just assume a successful remote piktc_svc kill following, and
	   set the pid1 to -1
	 */
	pid1 = -1;

	if (! (etcdiridp = lookupident("etcdir", MCRCFG)))
		logexit(OUTPUT, LOG_ERR, ERRID_IDENT_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupident() failure for %s", "etcdir");
	hostetcdir = etcdiridp->definition->component;

	/* kill the remote piktc_svc */
	if (inform) {
		printf("killing daemon (piktc_svc)...\n");
		fflush(stdout);
	}
	sprintf(cmd, "%s -x +C \"=psall | =grep piktc_svc | =grep -v grep | =awk \'{print \\$2}\' | =xargs =kill 2>/dev/null\" +H %s", piktx, host);
	system(cmd);
	/* wait for piktc_svc kill to finish */
	sleep(1+fork_delay);

	/* back up the current remote PIKT.conf */
	if (backup) {
		if (inform) {
			printf("backing up PIKT.conf...\n");
			fflush(stdout);
		}
		sprintf(cmd, "%s -x +C \"=cp -p =etcdir/PIKT.conf =etcdir/PIKT.conf.%s\" +H %s", piktx, bakext, host);
		system(cmd);
		/* wait for that to finish */
		sleep(1+fork_delay);
	}

	/* remote copy the revised PIKT.conf */
	if (inform) {
		printf("installing PIKT.conf...\n");
		fflush(stdout);
	}
	if (scpcmd)
		sprintf(cmd, "%s %s%s/PIKT.conf %s:%s/PIKT.conf", scpcmd, stgdir, hostetcdir, host, hostetcdir);
	else
		sprintf(cmd, "%s -x +C \"=scp %s%s/PIKT.conf %s:%s/PIKT.conf\" +H piktmaster", piktx, stgdir, hostetcdir, host, hostetcdir);
	system(cmd);
	/* wait for the scp to finish */
	sleep(1+fork_delay);
	if (inform) {
		/* add checks for this some other time? */
		printf("PIKT.conf %sinstalled\n", nullstring);
		fflush(stdout);
	}

	/* set mode & ownerships of remote PIKT.conf, and (re)start the
	   remote piktc_svc, all in one piktx command
	 */
	if (inform) {
		printf("restarting daemon (piktc_svc)...\n");
		fflush(stdout);
	}
	sprintf(cmd, "%s -x +C \"=chmod %s =etcdir/PIKT.conf; =chown %s =etcdir/PIKT.conf; =chgrp %s =etcdir/PIKT.conf; =bindir/piktc_svc\" +H %s", piktx, mode, filuid, filgid, host);
	system(cmd);
	/* wait for all that to finish */
	sleep(1+fork_delay);
	/* wait for the new piktc_svc to start up */
	sleep((unsigned int) restart_piktc_svc_wait);

	/* check for remote piktc_svc */
	pid2 = piktc_svc_pid(host, inform);

	if (isverboselog || isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "piktc_svc pid2 is %d", pid2);
	if ((pid2 < 0) /* || (pid1 == pid2) */ ) {
		if (inform) {
			printf("daemon not (re)started, or PIKT.conf improperly installed\n");
			fflush(stdout);
		}
	} else {
		if (inform) {
			printf("daemon (re)started\n");
			fflush(stdout);
		}
	}
	return;
}

/* kill either the alert daemon (piktd) or the service daemon (piktc_svc)
   on the given host
 */
void
kill_daemon (char *host, unsigned long service, int inform)
{
	static char fnnam[] = "kill_daemon";
	int kres;
	int *res;

	if (service == KILL_PIKTD_SERVICE)
		register_request(host, "kill_piktd", "daemon not killed", inform);
	else
		register_request(host, "kill_piktc_svc", "daemon not killed", inform);
	res = (int *) NULL;
	if (inform) {
		printf("killing daemon (%s)...\n", service == KILL_PIKTD_SERVICE ? "piktd" : "piktc_svc");
		fflush(stdout);
	}
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform) {
			clnt_pcreateerror(host);
			printf("daemon not killed\n");
			fflush(stdout);
		}
		register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
		return;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "kill_daemon (%s) request to host %s", service == KILL_PIKTD_SERVICE ? "piktd" : "piktc_svc", host);
	/* set the authunix credentials per the hostname, but with authuid
	   & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&reqserv.keysum.val, &reqserv.keysum.len, clnt);
	if (kres == 0) {
		if (service == KILL_PIKTD_SERVICE)
			res = kill_piktd_1((rqst *) &reqserv, clnt);
		else
			res = kill_piktc_svc_1((rqst *) &reqserv, clnt);
	}
	if (inform) {
		if (res) {
			switch (*res) {
				case ERROR:
					printf("daemon not killed\n");
					fflush(stdout);
					break;
				case NONE:
					printf("daemon not running\n");
					fflush(stdout);
					break;
				case UNAUTHORIZED:
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
					printf("daemon not killed\n");
					fflush(stdout);
					break;
				case DENIED:
					logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
					break;
				case 0:
					printf("daemon killed\n");
					fflush(stdout);
					break;
				default:
					break;
			}
		} else if (kres == 0) {
			if (service == KILL_PIKTC_SVC_SERVICE) {
				printf("daemon killed\n");
				fflush(stdout);
			} else {
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
			}
		}
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* restart either the alert daemon (piktd) or the service daemon (piktc_svc)
   on the given host
 */
void
restart_daemon (char *host, unsigned long service, int inform)
{
	static char fnnam[] = "restart_daemon";
	int pid1, pid2;
	struct timeval timeout;
	int kres;
	int *res;

	pid1 = pid2 = ERROR;
	if (service == RESTART_PIKTD_SERVICE)
		register_request(host, "restart_piktd", "daemon not (re)started", inform);
	else
		register_request(host, "restart_piktc_svc", "daemon not (re)started", inform);
	if (service == RESTART_PIKTC_SVC_SERVICE) {
		pid1 = piktc_svc_pid(host, inform);
		if (isverboselog || isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "piktc_svc pid1 is %d", pid1);
		if (pid1 < 0) {
			if (inform) {
				printf("daemon not (re)started\n");
				fflush(stdout);
			}
			register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
			return;
		}
	}
	res = (int *) NULL;
	if (inform) {
		printf("(re)starting daemon (%s)...\n", service == RESTART_PIKTD_SERVICE ? "piktd" : "piktc_svc");
		fflush(stdout);
	}
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform) {
			clnt_pcreateerror(host);
			printf("daemon not (re)started\n");
			fflush(stdout);
		}
		register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
		return;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "restart_daemon (%s) request to host %s", service == RESTART_PIKTD_SERVICE ? "piktd" : "piktc_svc", host);
	/* set the authunix credentials per the hostname, but with authuid
	   & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&reqserv.keysum.val, &reqserv.keysum.len, clnt);
	if (kres == 0) {
		if (service == RESTART_PIKTD_SERVICE) {
			res = restart_piktd_1((rqst *) &reqserv, clnt);
		} else /* if (service == RESTART_PIKTC_SVC_SERVICE) */ {
			/* we reset the timeout, else we might wait the
			   default timeout period for the now-dead slave
			   piktc_svc to respond, but of course it never does
		 	*/
			timeout.tv_sec = 0L;
			timeout.tv_usec = 0L;
			if (! clnt_control(clnt, CLSET_TIMEOUT, (char *)&timeout))
				logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set timeout");
			/* set retry timeout, but for udp only */
			if (! istcponly) {
				timeout.tv_sec = 2L;	/* should be higher
							   than the
							   CLSET_TIMEOUT
							   figures above
							 */
				timeout.tv_usec = 0L;	/* probably
							   redundant
							 */
				if (! clnt_control(clnt, CLSET_RETRY_TIMEOUT, (char *)&timeout))
					logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set retry timeout");
			} /* else if (istcponly)
				is not udp, and setting the retry timout doesn't apply
			   */
			res = restart_piktc_svc_1((rqst *) &reqserv, clnt);
		}
	}
	if (inform) {
		if (res) {
			switch (*res) {
				case ERROR:
				case NONE:
					printf("daemon not (re)started\n");
					fflush(stdout);
					break;
				case UNAUTHORIZED:
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
					printf("daemon not (re)started\n");
					fflush(stdout);
					break;
				case DENIED:
					logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
					break;
				case 0:
					if (service == RESTART_PIKTD_SERVICE) {
						printf("daemon (re)started\n");
						fflush(stdout);
					}
					break;
				default:
					break;
			}
		} else if (kres == 0) {
			if (service == RESTART_PIKTD_SERVICE) {
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
			}
		}
	}
	if (service == RESTART_PIKTC_SVC_SERVICE) {
		/* if the first piktc_svc died, then res should be the null
		   pointer; else it didn't die, so we got a return result,
		   and we infer restart failure
		 */
		if (! res) {
			/* we need to call clnt_destroy() here, because we
			   will be calling clnt_create() in piktc_svc_pid()
			   following
			 */
			/* auth_destroy(clnt->cl_auth); */
			clnt_destroy(clnt);
			/* wait for the new piktc_svc to start up */
			sleep((unsigned int) restart_piktc_svc_wait);
			pid2 = piktc_svc_pid(host, inform);
			if (isverboselog || isdebug)
				logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "piktc_svc pid2 is %d", pid2);
			if ((pid2 < 0) || (pid1 == pid2)) {
				if (inform) {
					printf("daemon not (re)started\n");
					fflush(stdout);
				}
			} else {
				if (inform) {
					printf("daemon (re)started\n");
					fflush(stdout);
				}
			}
		} else {
			if (inform) {
				printf("daemon not (re)started\n");
				fflush(stdout);
			}
		}
	}
	/* auth_destroy(clnt->cl_auth); */
	if (service == RESTART_PIKTD_SERVICE) {
		clnt_destroy(clnt);
	} else if (service == RESTART_PIKTC_SVC_SERVICE) {
		;	/* do nothing; we called clnt_destroy() before
			   the piktc_svc_pid() call above
			 */
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* report the status of each alert in the supplied list on the given host
 */
void
show_alerts_stata (char *host, nam *alts, int inform)
{
	nam *np;
	int res;

	register_request(host, "show_alert_status", "alert stata not shown", inform);
	if (inform) {
		printf("showing alert stata...\n");
		fflush(stdout);
	}
	foreach (np, alts) {
		if (np->state == ERROR) {
			if (inform) {
				printf("%s status not shown, unknown alert\n", np->name);
				fflush(stdout);
			}
			continue;
		}
		res = show_alert_status(host, np->name, nullstring, inform);
		if ((res == NONE) || (res == UNAUTHORIZED)) {
			if (inform) {
				printf("alert stata not shown\n");
				fflush(stdout);
			}
			break;
		}
		if (inform) {
			printf("%s %sactive\n", np->name, res == 1 ? nullstring : "not ");
			fflush(stdout);
		}
	}
	register_request(nullstring, nullstring, nullstring, FALSE);  /* deregister */
	return;
}

/* enable the given alert on the given host
 */
int
enable_alert (char *host, char *name, char *aux, int inform)
{
	static char fnnam[] = "enable_alert";
	int kres;
	int *res;

	res = (int *) NULL;
	alert.name = name;
	alert.aux = aux;
	alert.state = ADD;
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "edit_piktd_conf request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&alert.keysum.val, &alert.keysum.len, clnt);
	if (kres == 0) {
		res = edit_piktd_conf_1((ritm *) &alert, clnt);
	}
	if (res) {
		switch (*res) {
			case UNAUTHORIZED:
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				fflush(stderr);
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			default:
				break;
		}
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (res) {
		RETURN_RES;
	} else if (kres != 0) {
		RETURN_KRES;
	} else {
		return NONE;
	}
}

/* disable the given alert on the given host
 */
int
disable_alert (char *host, char *name, const char *aux, int inform)
{
	static char fnnam[] = "disable_alert";
	int kres;
	int *res;

	res = (int *) NULL;
	alert.name = name;
	alert.aux = (char *) aux;
	alert.state = DELETE;
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "edit_piktd_conf request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&alert.keysum.val, &alert.keysum.len, clnt);
	if (kres == 0) {
		res = edit_piktd_conf_1((ritm *) &alert, clnt);
	}
	if (res) {
		switch (*res) {
			case UNAUTHORIZED:
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			default:
				break;
		}
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (res) {
		RETURN_RES;
	} else if (kres != 0) {
		RETURN_KRES;
	} else {
		return NONE;
	}
}

/* execute the given alert on the given host
 */
int
execute_command (char *host, char *name, char *aux, int inform)
{
	static char fnnam[] = "execute_command";
	struct timeval timeout;
	char *execcmdout;
	unsigned int execcmdoutlen;
	char *keysan;
	int kres;
	execres *res;

	res = (execres *) NULL;
	if (isexec)
		clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	else if (isexecwait)
		clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, "tcp");
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "execute_command request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&command.keysum.val, &command.keysum.len, clnt);
	if (kres == 0) {
		command.name = mystrdup("execute_command", fnnam);
		command.aux = aux;
		if (isexec)
			command.state = EXECUTE;
		else if (isexecwait)
			command.state = EXECUTEWAIT;
		else 
			logexit(PIKTLOG, LOG_ERR, ERRID_ISEXECTYPE_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "isexec or isexecwait not set");
		if (isexec)
			timeout.tv_sec = 1L;
		else if (isexecwait)
			timeout.tv_sec = (long) rpc_timeout;
		timeout.tv_usec = 0L;
		if (! clnt_control(clnt, CLSET_TIMEOUT, (char *)&timeout))
			logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set timeout");
		if (isexec) {
			/* set retry timeout, but for exec/udp only */
			if (! istcponly) {
				timeout.tv_sec = 2L;
				timeout.tv_usec = 0L;	/* probably
							   redundant
							 */
				if (! clnt_control(clnt, CLSET_RETRY_TIMEOUT, (char *)&timeout))
					logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set retry timeout");
			} /* else if (istcponly)
			     is not udp, and setting the retry timout
			     doesn't apply
			   */
			res = (execres *) execute_command_1((ritm *) &command, clnt);
		} else if (isexecwait) {
			res = execute_commandwait_1((ritm *) &command, clnt);
		}
	}
	free(command.name);
	if (isexecwait) {
		if (res) {
			switch (res->result) {
				case UNAUTHORIZED:
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
					break;
				case DENIED:
					logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
					break;
				default:
					switch (data_encryption_type) {
						case 0:	/* no encryption */
							execcmdout = res->execres_u.execcmdout.txttype_val;
							execcmdoutlen = res->execres_u.execcmdout.txttype_len;
							break;
						case BLOWFISH_ENCRYPTION:
							bf_decrypt(&bc, res->execres_u.execcmdout.txttype_val, res->execres_u.execcmdout.txttype_len, &execcmdout, &execcmdoutlen);
							break;
						default:
							logexit(PIKTLOG, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid data_encryption_type %d", data_encryption_type);
							break;
					}
					if (isdebug) {
						keysan = sanitize(key, keylen);
						logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "key (keysan) is %s", /* key */ keysan);
						free(keysan);
					}
					if (execcmdout) {
						/* not sure why this is needed,
						   but it seems to be, because
						   we were seeing, in the no
						   encryption case, '@'
						   appended to the output
						 */
						*(execcmdout+execcmdoutlen) = '\0';
						fprintf(stdout, "%s", execcmdout);
						fflush(stdout);
					}
					switch (data_encryption_type) {
						case BLOWFISH_ENCRYPTION:
							free(execcmdout);
							break;
						case 0:	/* no encryption */
						default:
							break;
					}
					break;
			}
		} else {
			/* if you encounter the RPC difficulties described in
			   README.AIX, uncomment the '#ifdef AIX ...' workaround
			   following
			 */
			/*
			#ifdef AIX
			;
			#else
			 */
			logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
			/*
			#endif
			 */
		}
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (isexecwait && res) {
		RETURN_RES_RESULT;
	} else if (kres != 0) {
		RETURN_KRES;
	} else /* if (isexec || ! res) */ {
		/* if you encounter the RPC difficulties described in
		   README.AIX, uncomment the '#ifdef AIX ...' workaround
		   following
		 */
		/*
		#ifdef AIX
		if (! res)
			return TRUE;
		else
			return NONE;
		#else
		 */
		return NONE;
		/*
		#endif
		 */
	}
}

/* execute the given alert on the given host
 */
int
execute_alert (char *host, char *name, char *aux, int inform)
{
	static char fnnam[] = "execute_alert";
	struct timeval timeout;
	int kres;
	int *res;

	res = (int *) NULL;
	alert.name = name;
	alert.aux = aux;
	if (isexec)
		alert.state = EXECUTE;
	else if (isexecwait)
		alert.state = EXECUTEWAIT;
	else
		logexit(PIKTLOG, LOG_ERR, ERRID_ISEXECTYPE_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "isexec or isexecwait not set");
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "execute_alert request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&alert.keysum.val, &alert.keysum.len, clnt);
	if (kres == 0) {
		if (isexec)
			timeout.tv_sec = 1L;
		else if (isexecwait)
			timeout.tv_sec = (long) rpc_timeout;
		timeout.tv_usec = 0L;
		if (! clnt_control(clnt, CLSET_TIMEOUT, (char *)&timeout))
			logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set timeout");
		/* whether isexec or isexecwait, and is udp, set the retry timeout */
		if (! istcponly) {	/* hence is udp */
			if (isexec)
				timeout.tv_sec = 2L;
			else if (isexecwait)
				timeout.tv_sec = (long) rpc_retry_timeout;
			timeout.tv_usec = 0L;	/* probably redundant */
			if (! clnt_control(clnt, CLSET_RETRY_TIMEOUT, (char *)&timeout))
				logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set retry timeout");
		}
		res = execute_alert_1((ritm *) &alert, clnt);
	}
	if (isexecwait) {
		if (res) {
			switch (*res) {
				case UNAUTHORIZED:
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
					break;
				case DENIED:
					logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
					break;
				default:
					break;
			}
		} else {
			logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
		}
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (isexecwait && res) {
		RETURN_RES;
	} else if (kres != 0) {
		RETURN_KRES;
	} else /* if (isexec || ! res) */ {
		return NONE;
	}
}

/* install the given file on the given host
 */
int
install_file (char *host, char *name, char *dir, const char *ext, char *mode, char *filuid, char *filgid, int backup, int inform)
{
	static char fnnam[] = "install_file";
	int isdefault;
	static nam *np;
	static char *fn;
	struct stat fbuf;
	off_t len;
	FILE *fp;
	int kres;
	char *filtext;
	char *keysan;
	int *res;

	res = (int *) NULL;
	isdefault = TRUE;	/* assume the default dir */
	if (! *dir) {	/* signals that this is an other file
			   install, in which case override
			   the default dir, mode, filuid & filgid
			 */
		if (! (np = lookupname(files, name))) {
			if (! filesall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		isdefault = FALSE;
		_file.dir = np->dir;
		_file.file = np->file;
		_file.ext = (char *) ext;
		_file.mode = np->mode;
		_file.filuid = np->filuid;
		_file.filgid = np->filgid;
	} else if (! strcmp(ext, "NIL")) {   /* signals that this is a program
				               install, in which case possibly
				               override the default dir, mode,
				               filuid & filgid
				             */
		if (! (np = lookupname(programs, name))) {
			if (! programsall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		if (np->dir && *(np->dir)) {
			isdefault = FALSE;
			_file.dir = np->dir;
			_file.file = np->file;
		} else {
			_file.dir = dir;
			_file.file = name;
		}
		_file.ext = (char *) ext;
		_file.mode = np->mode;
		_file.filuid = np->filuid;
		_file.filgid = np->filgid;
	} else {			/* go with the defaults */
		_file.dir = dir;
		_file.file = name;
		_file.ext = (char *) ext;
		_file.mode = mode;
		_file.filuid = filuid;
		_file.filgid = filgid;
	}
	if (isdefault) {
		fn = (char *) mymalloc((strlen(stgdir)+strlen(_file.file)+strlen(_file.ext)+2) * sizeof(char), fnnam);
		sprintf(fn, "%s/%s%s", stgdir, _file.file, strcmp(_file.ext, "NIL") ? _file.ext : nullstring);
	} else {
		fn = (char *) mymalloc((strlen(stgdir)+strlen(_file.dir)+strlen(_file.file)+strlen(_file.ext)+2) * sizeof(char), fnnam);
		sprintf(fn, "%s%s/%s%s", stgdir, _file.dir, _file.file, strcmp(_file.ext, "NIL") ? _file.ext : nullstring);
	}
	if (stat(fn, &fbuf)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "stat() failure for %s", fn);
		free(fn);
		return 0;
	}
	len = fbuf.st_size;
	_file.len = len;
	filtext = (char *) mymalloc((size_t)(len+1)*sizeof(char), fnnam);
	if ((fp = fopen(fn, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", fn);
		free(fn);
		return 0;
	}
	if (fread(filtext, sizeof(char), (size_t)len, fp) == 0) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_READ_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fread() error for %s", fn);
		free(fn);
		return 0;
	}
	if (fclose(fp)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", fn);
		free(fn);
		return 0;
	}
	free(fn);
	*(filtext+len) = '\0';
	_file.backup = backup;
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, "tcp");
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "install_file request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&_file.keysum.val, &_file.keysum.len, clnt);
	if (kres == 0) {
		switch (data_encryption_type) {
			case 0:	/* no encryption */
				_file.text.val = filtext;
				_file.text.len = _file.len;
				break;
			case BLOWFISH_ENCRYPTION:
				bf_encrypt(&bc, filtext, _file.len, &_file.text.val, &_file.text.len);
				break;
			default:
				logexit(OUTPUT, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid data_encryption_type %d", data_encryption_type);
				break;
		}
		if (isdebug) {
			keysan = sanitize(key, keylen);
			logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "key (keysan) is %s", /* key */ keysan);
			free(keysan);
		}
		res = install_file_1((rfil *) &_file, clnt);
	}
	if (res) {
		switch (*res) {
			case UNAUTHORIZED:
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			default:
				break;
		}
	} else {
		/* if you encounter the RPC difficulties described in
		   README.AIX, uncomment the '#ifdef AIX ...' workaround
		   following
		 */
		/*
		#ifdef AIX
		;
		#else
		 */
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
		/*
		#endif
		 */
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	FREEVOID(_file.text.val);
	switch (data_encryption_type) {
		case BLOWFISH_ENCRYPTION:
			free(filtext);
			break;
		case 0:	/* no encryption */
		default:
			break;
	}
	if (res) {
		RETURN_RES;
	} else if (kres != 0) {
		RETURN_KRES;
	} else {
		/* if you encounter the RPC difficulties described in
		   README.AIX, uncomment the '#ifdef AIX ...' workaround
		   following
		 */
		/*
		#ifdef AIX
		return TRUE;
		#else
		 */
		return NONE;
		/*
		#endif
		 */
	}
}

/* fetch the given file from the given host
 */
int
fetch_file (char *host, char *name, char *dir, const char *ext, int inform)
{
	static char fnnam[] = "fetch_file";
	int isdefault;
	static nam *np;
	char *fdir, *ffile, *fext;
	char *sfn, *tfn;
	struct stat fbuf;
	char *cmd;
	struct timeval timeout;
	char *execcmdout;
	unsigned int execcmdoutlen;
	char *keysan;
	FILE *tout;
	int kres;
	execres *res;

	execcmdout = (char *) NULL;
	cmd = (char *) NULL;
	res = (execres *) NULL;
	isdefault = TRUE;	/* assume the default dir */
	fdir = (char *) NULL;
	if (! *dir) {	/* signals that this is an other file
			   fetch
			 */
		if (! (np = lookupname(files, name))) {
			if (! filesall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		isdefault = FALSE;
		fdir = np->dir;
		ffile = np->file;
	} else if (! strcmp(ext, "NIL")) {   /* signals that this is a program
				               fetch
				             */
		if (! (np = lookupname(programs, name))) {
			if (! programsall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		if (np->dir && *(np->dir)) {
			isdefault = FALSE;
			fdir = np->dir;
			ffile = np->file;
		} else {
			fdir = dir;
			ffile = name;
		}
	} else {			/* go with the defaults */
		fdir = dir;
		ffile = name;
	}
	if (hostmatch(host, master))
		fdir = getlocdir(fdir);
	fext = (char *) ext;
	sfn = (char *) mymalloc((strlen(fdir)+strlen(ffile)+strlen(fext)+2) * sizeof(char), fnnam);
	sprintf(sfn, "%s/%s%s", fdir, ffile, strcmp(fext, "NIL") ? fext : nullstring);
	if (isdefault) {
		tfn = (char *) mymalloc((strlen(difdir)+strlen(ffile)+strlen(fext)+2) * sizeof(char), fnnam);
		sprintf(tfn, "%s/%s%s", difdir, ffile, strcmp(fext, "NIL") ? fext : nullstring);
	} else {
		tfn = (char *) mymalloc((strlen(difdir)+strlen(fdir)+strlen(ffile)+strlen(fext)+2) * sizeof(char), fnnam);
		sprintf(tfn, "%s%s/%s%s", difdir, fdir, ffile, strcmp(fext, "NIL") ? fext : nullstring);
	}
	if (hostmatch(host, master)) {
		if (stat(sfn, &fbuf)) {
			FREE(fdir);
			free(sfn);
			free(tfn);
			return FALSE;
		} else {
			cmd = (char *) mymalloc((strlen(cpcmd)+strlen(sfn)+strlen(tfn)+15) * sizeof(char), fnnam);
			sprintf(cmd, "%s %s %s 2>/dev/null", cpcmd, sfn, tfn);
			system(cmd);
			free(cmd);
			FREE(fdir);
			free(sfn);
			free(tfn);
			return TRUE;
		}
	} else {
		clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, "tcp");
		if (clnt == NULL) {
			if (inform)
				clnt_pcreateerror(host);
			free(sfn);
			free(tfn);
			return NONE;
		}
		logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "fetch_file request to host %s", host);
		/* set the authunix credentials per the hostname, but with
		   authuid & authgid set per the PIKT.conf settings;
		   this replaces:
		(AUTH *) clnt->cl_auth = authunix_create_default();
		 */
		(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
		kres = get_key(&command.keysum.val, &command.keysum.len, clnt);
		if (kres == 0) {
			if (! strcmp(fdir, ALTDIR) && ! strcmp(fext, ".alt")) {
				command.name = mystrdup("fetch_file_alerts", fnnam);
			} else if (! strcmp(fdir, OBJDIR) && ! strcmp(fext, ".obj")) {
				command.name = mystrdup("fetch_file_objects", fnnam);
			} else if ((! strcmp(fdir, PRGDIR) || ! strcmp(fext, "NIL"))) {
				command.name = mystrdup("fetch_file_programs", fnnam);
			} else if (! *fext) {
				command.name = mystrdup("fetch_file_files", fnnam);
			} /* else? */
			/* on any sane unix system, it should always be the
			   case that the cat command is "/bin/cat"
		 	 */
			command.aux = (char *) mymalloc((strlen("/bin/cat")+strlen(sfn)+14) * sizeof(char), fnnam);
			sprintf(command.aux, "%s %s 2>/dev/null", "/bin/cat", sfn);
			command.state = EXECUTEWAIT;
			timeout.tv_sec = (long) rpc_timeout;
			timeout.tv_usec = 0L;
			if (! clnt_control(clnt, CLSET_TIMEOUT, (char *)&timeout))
				logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set timeout");
			res = execute_commandwait_1((ritm *) &command, clnt);
			free(command.name);
			free(command.aux);
		}
		if (res) {
			switch (res->result) {
				case UNAUTHORIZED:
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
					break;
				case DENIED:
					logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
					break;
				default:
					switch (data_encryption_type) {
						case 0:	/* no encryption */
							execcmdout = res->execres_u.execcmdout.txttype_val;
							execcmdoutlen = res->execres_u.execcmdout.txttype_len;
							break;
						case BLOWFISH_ENCRYPTION:
							bf_decrypt(&bc, res->execres_u.execcmdout.txttype_val, res->execres_u.execcmdout.txttype_len, &execcmdout, &execcmdoutlen);
							break;
						default:
							logexit(OUTPUT, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid data_encryption_type %d", data_encryption_type);
							break;
					}
					if (isdebug) {
						keysan = sanitize(key, keylen);
						logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "key (keysan) is %s", /* key */ keysan);
						free(keysan);
					}
					if (execcmdout) {
						if ((tout = fopen(tfn, "w")) == NULL) {
							logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", tfn);
							res->result = ERROR;
						} else {
							fprintf(tout, "%s", execcmdout);
							if (fclose(tout)) {
								logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", tfn);
								res->result = ERROR;
							}
						}
					}
					break;
			}
		} else {
			/* if you encounter the RPC difficulties described in
			   README.AIX, uncomment the '#ifdef AIX ...' workaround
			   following
			 */
			/*
			#ifdef AIX
			;
			#else
			 */
			logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
			/*
			#endif
			 */
		}
		FREE(cmd);
		switch (data_encryption_type) {
			case BLOWFISH_ENCRYPTION:
				FREE(execcmdout);
				break;
			case 0:	/* no encryption */
			default:
				break;
		}
		free(sfn);
		free(tfn);
		/* auth_destroy(clnt->cl_auth); */
		clnt_destroy(clnt);
		if (res) {
			RETURN_RES_RESULT;
		} else if (kres != 0) {
			RETURN_KRES;
		} else {
			/* if you encounter the RPC difficulties described in
			   README.AIX, uncomment the '#ifdef AIX ...' workaround
			   following
			 */
			/*
			#ifdef AIX
			return TRUE;
			#else
			 */
			return NONE;
			/*
			#endif
			 */
		}
	}
}

/* report if the given file exists on the given host
 */
int
cksum_file (char *host, char *name, char *dir, const char *ext, int level, int inform)
{
	static char fnnam[] = "cksum_file";
	static nam *np;
	int kres;
	cksumres *res;

	struct stat fbuf;
	char *sfn;		/* source (i.e., staging) filename */
	char *scs;		/* source (i.e., staging) checksum string */
	int isstgfil;

	isstgfil = TRUE;
	res = (cksumres *) NULL;
	if (! *dir) {	/* signals that this is an other file
			   lookup, in which case override
			   the default dir
			 */
		if (! (np = lookupname(files, name))) {
			if (! filesall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		_file.dir = np->dir;
		_file.file = np->file;
	} else if (! strcmp(ext, "NIL")) {   /* signals that this is a program
				               lookup, in which case possibly
				               override the default dir
				             */
		if (! (np = lookupname(programs, name))) {
			if (! programsall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		if (np->dir && *(np->dir)) {
			_file.dir = np->dir;
			_file.file = np->file;
		} else {
			_file.dir = dir;
			_file.file = name;
		}
	} else {
		_file.dir = dir;
		_file.file = name;
	}
	_file.ext = (char *) ext;
	_file.mode = nullstring;
	_file.filuid = nullstring;
	_file.filgid = nullstring;
	_file.len = level;	/* set the checksum level, 1-5 */
	_file.text.val = nullstring;
	_file.text.len = 0;
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "cksum_file request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&_file.keysum.val, &_file.keysum.len, clnt);
	if (kres == 0) {
		res = (cksumres *) cksum_file_1((rfil *) &_file, clnt);
	}
	if (res) {
		switch (res->result) {
			case UNAUTHORIZED:
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			case NONE:
				logreturn(OUTPUT, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "%s%s not found", _file.file, strcmp(_file.ext,"NIL") ? _file.ext : nullstring);
				break;
			default:
				sfn = (char *) mymalloc((strlen(stgdir)+strlen((! *dir) || (! strcmp(ext, "NIL") && np->dir && *(np->dir)) ? np->dir : dir)+strlen(_file.file)+strlen(_file.ext)+2) * sizeof(char), fnnam);
				sprintf(sfn, "%s%s/%s%s", stgdir, ((! *dir) || (! strcmp(ext, "NIL") && np->dir && *(np->dir)) ? np->dir : nullstring), _file.file, strcmp(_file.ext, "NIL") ? _file.ext : nullstring);
				if (stat(sfn, &fbuf)) {	/* 0 signals success */
					logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_NOT_FOUND, __FILE__, __LINE__, fnnam, FALSE, "%s staging file not found", sfn);
				isstgfil = FALSE;
				} else {
					scs = checksumfil(level, sfn);
					if (isverboselog || isdebug) {
						logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "staging file checksum is %s", scs);
						logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "slave file checksum is %s", res->cksumres_u.cksum);
					}
					if (csstrcmp(scs, res->cksumres_u.cksum)) {
						printf("%s%s checksums differ\n", _file.file, strcmp(_file.ext, "NIL") ? _file.ext : nullstring);
					} else {
						/* this is getting to be too
						   wordy already
						printf("%s%s checksums match\n", _file.file, strcmp(_file.ext, "NIL") ? _file.ext : nullstring);
						 */
						;
					}
					fflush(stdout);
				}
				free(sfn);
				break;
		}
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (! isstgfil)
		return FALSE;
	if (res) {
		RETURN_RES_RESULT;
	} else if (kres != 0) {
		RETURN_KRES;
	} else {
		return NONE;
	}
}

/* delete the given file on the given host
 */
int
delete_file (char *host, char *name, char *dir, const char *ext, int backup, int inform)
{
	static char fnnam[] = "delete_file";
	static nam *np;
	int kres;
	int *res;

	res = (int *) NULL;
	if (! *dir) {	/* signals that this is an other file
			   deletion, in which case override
			   the default dir
			 */
		if (! (np = lookupname(files, name))) {
			if (! filesall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		_file.dir = np->dir;
		_file.file = np->file;
	} else if (! strcmp(ext, "NIL")) {   /* signals that this is a program
				               deletion, in which case possibly
				               override the default dir
				             */
		if (! (np = lookupname(programs, name))) {
			if (! programsall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		if (np->dir && *(np->dir)) {
			_file.dir = np->dir;
			_file.file = np->file;
		} else {
			_file.dir = dir;
			_file.file = name;
		}
	} else {
		_file.dir = dir;
		_file.file = name;
	}
	_file.ext = (char *) ext;
	_file.mode = nullstring;
	_file.filuid = nullstring;
	_file.filgid = nullstring;
	_file.len = 0;
	_file.text.val = nullstring;
	_file.text.len = 0;
	_file.backup = backup;
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "delete_file request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&_file.keysum.val, &_file.keysum.len, clnt);
	res = delete_file_1((rfil *) &_file, clnt);
	if (res) {
		switch (*res) {
			case UNAUTHORIZED:
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			default:
				break;
		}
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (res) {
		RETURN_RES;
	} else if (kres != 0) {
		RETURN_KRES;
	} else {
		return NONE;
	}
}

/* report if the given file exists on the given host
 */
int
list_file (char *host, char *name, char *dir, const char *ext, int status, int inform)
{
	static char fnnam[] = "list_file";
	static nam *np;
	int kres;
	int *res;

	res = (int *) NULL;
	if (! *dir) {	/* signals that this is an other file
			   lookup, in which case override
			   the default dir
			 */
		if (! (np = lookupname(files, name))) {
			if (! filesall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		_file.dir = np->dir;
		_file.file = np->file;
	} else if (! strcmp(ext, "NIL")) {   /* signals that this is a program
				               lookup, in which case possibly
				               override the default dir
				             */
		if (! (np = lookupname(programs, name))) {
			if (! programsall) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_NAME_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "lookupname() failure for %s", name);
				return 0;
			} else
				return IGNORE;
		}
		if (np->dir && *(np->dir)) {
			_file.dir = np->dir;
			_file.file = np->file;
		} else {
			_file.dir = dir;
			_file.file = name;
		}
	} else {
		_file.dir = dir;
		_file.file = name;
	}
	_file.ext = (char *) ext;
	_file.mode = nullstring;
	_file.filuid = nullstring;
	_file.filgid = nullstring;
	_file.len = 0;
	_file.text.val = nullstring;
	_file.text.len = 0;
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "list_file request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&_file.keysum.val, &_file.keysum.len, clnt);
	if (kres == 0) {
		res = list_file_1((rfil *) &_file, clnt);
	}
	if (res) {
		switch (*res) {
			case UNAUTHORIZED:
				if (status != IGNORE) {
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				}
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			default:
				break;
		}
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (res) {
		if (status == IGNORE) {
			switch (*res) {
				case 0:
					return TRUE;
					break;
				case UNAUTHORIZED:
					return IGNORE;
					break;
				default:
					return IGNORE;
					break;
			}
		} else {
			RETURN_RES;
		}
	} else if (kres != 0) {
		RETURN_KRES;
	} else {
		return NONE;
	}
}

/* report the status of the given alert on the given host
 */
int
show_alert_status (char *host, char *name, const char *aux, int inform)
{
	static char fnnam[] = "show_alert_status";
	int kres;
	int *res;

	res = (int *) NULL;
	alert.name = name;
	alert.aux = (char *) aux;
	alert.state = IGNORE;
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(host);
		return NONE;
	}
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "show_alert_status request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&alert.keysum.val, &alert.keysum.len, clnt);
	if (kres == 0) {
		res = show_alert_status_1((ritm *) &alert, clnt);
	}
	if (res) {
		switch (*res) {
			case UNAUTHORIZED:
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			default:
				break;
		}
	} else if (kres == 0) {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (res) {
		RETURN_RES;
	} else if (kres != 0) {
		RETURN_KRES;
	} else {
		return NONE;
	}
}

/* returns the piktc_svc pid of the given remote host
 */
int
piktc_svc_pid (char *host, int inform)
{
	static char fnnam[] = "piktc_svc_pid";
	int *res;
	int kres;
	int pid;

	/* we retain the previous restart_piktc_svc register_request() */
	res = (int *) NULL;
	if (inform) {
		printf("determining piktc_svc pid...\n");
		fflush(stdout);
	}
	clnt = clnt_create(host, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform) {
			clnt_pcreateerror(host);
			printf("piktc_svc pid not determined\n");
			fflush(stdout);
		}
		/* because we don't do a local register_request() above
		register_request(nullstring, nullstring, nullstring, FALSE);
		 */
		return ERROR;
	}
	if (isverboselog || isdebug)
		logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "piktc_svc_pid request to host %s", host);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	kres = get_key(&reqserv.keysum.val, &reqserv.keysum.len, clnt);
	if (kres == 0)
		res = piktc_svc_pid_1((rqst *) &reqserv, clnt);
	if (inform) {
		if (res) {
			switch (*res) {
				case ERROR:
					printf("piktc_svc pid not determined: possibly multiple piktc_svc running\n");
					fflush(stdout);
					break;
				case NONE:
					printf("piktc_svc pid not determined: pid unavailable\n");
					fflush(stdout);
					break;
				case UNAUTHORIZED:
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
					break;
				case DENIED:
					logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
					break;
				default:
					break;
			}
		} else if (kres == 0) {
			logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
		}
	}
	if (res)
		pid = *res;
	else
		pid = ERROR;
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	/* because we don't do a local register_request() above
	register_request(nullstring, nullstring, nullstring, FALSE);
	 */
	return pid;
}

/* do some necessary initializations; then, if we are not doing call_back
   (it's on by default, so this implies that we have set call_back in
   PIKT.conf to FALSE), just return; else (we are doing call_back, so)
   register a service request with the local piktc_svc
 */
void
register_request (char *host, char *request, char *failmsg, int inform)
{
	static char fnnam[] = "register_request";
	struct timeval timeout;
	int *res;

	/* we need to set reqserv.hostname & reqserv.request for subsequent
	   get_key() call and other calls
	 */
	reqserv.hostname = host;
	reqserv.request = request;
	reqserv.state = NONE;
	if (! IS_CALL_BACK)
		return;
	res = (int *) NULL;
	if (! *request) {	/* request is nullstring, i.e., a deregistration */
		if (isexec)
			reqserv.state = EXECUTE;	/* leave in place for
							   possible late
							   callbackers
							 */
		/* now done as an initialization just above:
		else
			reqserv.state = NONE;
		 */
	} else {
		if (isexec)
			reqserv.state = EXECUTE;
		else if (isexecwait)
			reqserv.state = EXECUTEWAIT;
	}
	/* hostname is the name of the local host--i.e., the machine
	   where the piktc service request is issuing from
	 */
	clnt = clnt_create(hostname, PIKTCPROG, PIKTCVERS, transports[istcponly]);
	if (clnt == NULL) {
		if (inform)
			clnt_pcreateerror(hostname);
		return;
	}
	if (isverboselog || isdebug)
		logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "register_request request to host %s", hostname);
	/* set the authunix credentials per the hostname, but with authuid		  & authgid set per the PIKT.conf settings; this replaces:
	(AUTH *) clnt->cl_auth = authunix_create_default();
	 */
	(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
	/* we don't call get_key() here */
	/* if the master piktc_svc is busy handling many incoming callback
	   requests for an exec (no wait), we need to up the timeout; else
	   go with the default
	 */
	if (isexec) {
		timeout.tv_sec = (long) rpc_timeout;
		timeout.tv_usec = 0L;
		if (! clnt_control(clnt, CLSET_TIMEOUT, (char *)&timeout))
			logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set timeout");
		if (! istcponly) {	/* hence is udp */
			timeout.tv_sec = (long) rpc_retry_timeout;
			timeout.tv_usec = 0L;	/* probably redundant */
			if (! clnt_control(clnt, CLSET_RETRY_TIMEOUT, (char *)&timeout))
				logreturn(PIKTLOG, LOG_WARNING, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_control() failure to set retry timeout");
		}
	}
	res = register_request_1((rqst *) &reqserv, clnt);
	if (inform) {
		if (res) {
			switch (*res) {
				case UNAUTHORIZED:
					logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
					break;
				case DENIED:
					logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
					break;
				default:	/* should only be 0, i.e., success */
					break;
			}
		} else {
			logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from local server");
		}
	}
	/* auth_destroy(clnt->cl_auth); */
	clnt_destroy(clnt);
	if (res && (*res == 0)) {
		return;
	} else {
		if (inform) {
			printf("%s\n", failmsg);
			fflush(stdout);
		}
		remove(piktclock);
		exit(exstat);
	}
}

/* get the encrypted authentication/session key from the slave piktc_svc,
   decrypt it, and compute its MD5 checksum
 */
int
get_key (char **kysm, unsigned int *kyln, CLIENT *kclnt)
{
	static char fnnam[] = "get_key";
	unsigned int keyenclen;
	char *keysan;
	char *keyencsan;
	keyencres *kres;

	reqserv.state = endian;
	reqserv.enctyp = data_encryption_type;
	/* this appears to cause problems:
	reqserv.keysum.val = (char *) NULL;
	reqserv.keysum.len = 0;
	 */
	kres = (keyencres *) get_key_1((rqst *) &reqserv, kclnt);
	if (kres) {
		switch (kres->result) {
			case UNAUTHORIZED:
				logreturn(OUTPUT, LOG_ERR, ERRID_SERVICE_UNAUTHORIZED, __FILE__, __LINE__, fnnam, FALSE, "service unauthorized");
				return UNAUTHORIZED;
				break;
			case DENIED:
				logreturn(OUTPUT, LOG_ERR, ERRID_ACCESS_DENIED, __FILE__, __LINE__, fnnam, FALSE, "access denied");
				break;
			case ERROR:
				logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_ENCRYPTION_MISMATCH, __FILE__, __LINE__, fnnam, FALSE, "data_encryption_type mismatch");
				return ERROR;
				break;
			case 0:
				FREE(keyenc);
				keyenc = (char *) mycalloc((kres->keyencres_u.keyenc.txttype_len)+1, sizeof(char), fnnam);
				memcpy(keyenc, kres->keyencres_u.keyenc.txttype_val, kres->keyencres_u.keyenc.txttype_len);
				keyenclen = kres->keyencres_u.keyenc.txttype_len;
				FREE(key);
				/* we don't check the return value of deckey(),
				   assuming that enckey() on the slave
				   piktc_svc would only return a non-weak key
				 */
				if (auth_encryption_type != 0)
					deckey(kres->keyencres_u.keyenc.txttype_val, kres->keyencres_u.keyenc.txttype_len, &key, &keylen);
				else {
					key = (char *) mycalloc((kres->keyencres_u.keyenc.txttype_len)+1, sizeof(char), fnnam);
                                        memcpy(key, kres->keyencres_u.keyenc.txttype_val, kres->keyencres_u.keyenc.txttype_len);
					keylen = kres->keyencres_u.keyenc.txttype_len;
				}
				FREE(keysum);
				keysum = checksumstr(5, key);
				*kyln = strlen(keysum);
                                *kysm = mycalloc((*kyln)+1, sizeof(char), fnnam);
                                memcpy(*kysm, keysum, *kyln);
				if (isdebug) {
					keysan = sanitize(key, keylen);
					keyencsan = sanitize(keyenc, keyenclen);
					logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "key (keysan) is %s, keysum is %s, keyenc (keyencsan) is %s, keyenclen is %d", /* key */ keysan, keysum, /* keyenc */ keyencsan, keyenclen);
					free(keysan);
					free(keyencsan);
				}
				return 0;
				break;
		}
	} else {
		logreturn(OUTPUT, LOG_ERR, ERRID_SERVER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from server");
		return NONE;
	}
	return NONE;
}

/* decrypt the auth/session key
 */
int
deckey (char *ke, unsigned int kelen, char **ky, unsigned int *kylen)
{
	static char fnnam[] = "deckey";
	int res;

	res = 0;
	bf_setkey(&bc, pvtkey, pvtkeylen);
	bf_decrypt(&bc, ke, kelen, ky, kylen);
	switch (data_encryption_type) {
		case 0:	/* no encryption */
			break;
		case BLOWFISH_ENCRYPTION:
			res = bf_setkey(&bc, *ky, *kylen);
			break;
		default:
			logexit(PIKTLOG, LOG_ERR, ERRID_DATA_ENCRYPTION_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "invalid data_encryption_type %d", data_encryption_type);
			break;
	}
	return res;
}

#else	/* #ifndef PIKTC_SVC */		/* piktc_svc */

/* call back to the master piktc_svc, to see if this hostname and service
   request are registered in that piktc_svc
 */
int
call_back (char *host, const char *request, int inform)
{

	static char fnnam[] = "call_back";

#ifndef NO_RPC_SVC

	/* for systems supporting the call_back feature */

	int *res;

	if (IS_CALL_BACK) {

		res = (int *) NULL;
		reqserv.hostname = host;
		reqserv.request = (char *) request;
		reqserv.state = IGNORE;

		clnt = clnt_create(master, PIKTCPROG, PIKTCVERS, transports[istcponly]);
		if (clnt == NULL) {
			/*
			clnt_pcreateerror(master);
			*/
			logreturn(PIKTLOG, LOG_ERR, ERRID_CLNT_CONTROL_ERROR, __FILE__, __LINE__, fnnam, FALSE, "clnt_create() error for host %s", master);
			return DENIED;
		}
		if (isverboselog || isdebug)
			logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "call_back request to host %s", master);
		/* set the authunix credentials per the hostname, but with
		   authuid & authgid set per the PIKT.conf settings;
		   this replaces:
		   (AUTH *) clnt->cl_auth = authunix_create_default();
		 */
		(AUTH *) clnt->cl_auth = authunix_create(hostname, authuid, authgid, 1, &authgid);
		/* we don't call get_key() here */
		res = call_back_1((rqst *) &reqserv, clnt);
		if (res) {
			switch (*res) {
				case DENIED:
					logreturn(PIKTLOG, LOG_ERR, ERRID_SERVICE_REQUEST_NOT_REGISTERED_ON_MASTER, __FILE__, __LINE__, fnnam, FALSE, "%s request not registered on master %s", request, master);
					break;
				default:	/* should only be 0, i.e., success */
					if (isverboselog)
						logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s request registered on master %s", request, master);
					break;
			}
		} else
			logreturn(PIKTLOG, LOG_ERR, ERRID_MASTER_UNRESPONSIVE, __FILE__, __LINE__, fnnam, FALSE, "no response from master server");
		/* auth_destroy(clnt->cl_auth); */
		clnt_destroy(clnt);
		if (res) {
			switch (*res) {
				case DENIED:
					return DENIED;
					break;
				default:
					return 0;
					break;
			}
		} else {
			return DENIED;
		}

	} else {

		return 0;	/* if call_back has been turned off,
				   always return success; note that if
				   call_back has been turned off, we
				   should never enter this function anyway
				 */

	}

#else	/* #ifndef NO_RPC_SVC */

	/* function stub for systems not supporting call_back and where
	   NO_RPC_SVC is #define'd; this is required so that the
	   svc_rpc_clnt.c compile doesn't come up empty-handed (i.e.,
	   that there is a resulting svc_rpc_clnt.o for linking into
	   piktc_svc)
	 */

	return 0;

#endif	/* #ifndef NO_RPC_SVC */

}

#endif	/* #ifndef PIKTC_SVC */
