%{

/*
   hst.y -- PIKT history file parser

   $Header: /cvs/PIKT/pikt/src/piktd/hst.y,v 1.26 2002/08/23 17:00:24 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"
#include "piktd.h"

#define NOT_HSTERR	(alarmcurr && ! HSTERR)

char _hstparse_[] = "hstparse";

vl *vlp;

int k;

/* in input:, this case is suspicious:

		| vals log

   i don't understand under what circumstances that .hst file sequence would
   occur; i leave it in, however, because if it's there, it probably occurred
   sometime in the past, and removing it might lead to a parser error--but
   maybe it should!
 */

%}

%union {
	char *sval;
	int ival;
	long lval;
}

%token _ALM _VAR _KEY _LOG _VAL
%token <ival> _INT
%token <lval> _LNG
%token <sval> _KST _STR

%%

start:		alms {
			if (alarmcurr && KEYS && keycurr) {
				logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_NEW_KEY_ADDED, __FILE__, __LINE__, _hstparse_, FALSE, "in %s, new key(s) added", ALMNAM);
				HSTERR = TRUE;
			}
		}
		;

alms:		alm
		| alms alm
		;

alm:		_ALM _STR {
			/* for the previous alarm (if any so far), if we have
			   keys, but not all of them were seen in the
			   history file (we didn't make it all the way through
			   the list), one or more keys have ben added, and
			   we signify error
			 */
			if (alarmcurr && KEYS && keycurr) {
				logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_NEW_KEY_ADDED, __FILE__, __LINE__, _hstparse_, FALSE, "in %s, new key(s) added", ALMNAM);
				HSTERR = TRUE;
			}
			if ((alarmcurr = lookupalarm($2))) {
				keycurr = KEYS;
				valcurr = VALHST;
				HSTERR = FALSE;
			} else {
				logreturn(PIKTLOG, LOG_WARNING, ERRID_HISTORY_FILE_ALARM_RETIRED, __FILE__, __LINE__, _hstparse_, FALSE, "in %s, alarm in history file now retired or inactive", $2);
			}
			free($2);
		}
		input
		;

input:		log
		| vars
		| vars log
		| vars vals
		| vars vals log
		| vars keys
		| vars keys log
		| vars keys vals
		| vars keys vals log
		| vals log
		;


vars:		var
		| vars var
		;

var:		_VAR _INT _INT _INT _INT _STR {
			if (NOT_HSTERR) {
				verifyvar($2, $6);
				verifydim($2, $4);
				verifytyp($2, $5);
			}
			free($6);
		}
		;

keys:		key
		| keys key
		;

key:		_KEY _INT _INT {
			if (NOT_HSTERR) {
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = 0L;
				verifykey($2, arndx, (char *)NULL);
			}
		}
		| _KEY _INT _INT _INT {			/* 1 dimension */
			if (NOT_HSTERR) {
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = 0L;
				arndx[1] = (signed long) $4;
				verifykey($2, arndx, (char *)NULL);
			}
		}
		| _KEY _INT _INT _INT _INT {		/* 2 dimensions */
			if (NOT_HSTERR) {
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = 0L;
				arndx[1] = (signed long) $4;
				arndx[2] = (signed long) $5;
				verifykey($2, arndx, (char *)NULL);
			}
		}
		| _KEY _INT _INT _INT _INT _INT {	/* 3 dimensions */
			if (NOT_HSTERR) {
				arndx[0] = 0L;
				arndx[1] = (signed long) $4;
				arndx[2] = (signed long) $5;
				arndx[3] = (signed long) $6;
				verifykey($2, arndx, (char *)NULL);
			}
		}
		| _KEY _INT _KST {
			if (NOT_HSTERR) {
				/* pass arndxnil instead
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = NILARNDX;
				 */
				verifykey($2, arndxnil, $3);
			}
			free($3);
		}
		;

vals:		val
		| vals val
		;

val:		_VAL _INT _KST _STR {
			if (NOT_HSTERR) {
				/* pass arndxnil instead
				for (k=0; k<=MAXARYDIM; k++)
					arndx[k] = NILARNDX;
				 */
				verifyval($2, arndx, $3);
				if (! (vlp = regvl(VALHST, $2, arndxnil, $3, STRING))) {
					logreturn(PIKTLOG, LOG_WARNING, ERRID_REGVL_ERROR, __FILE__, __LINE__, _hstparse_, FALSE, "in %s, regvl() error for var %s, key %s, val %s", ALMNAM, VARNAM($2), $3, $4);
					HSTERR = TRUE;
				} else {
					vlp->str = $4;
				}
			}
			free($3);
		}
		;

log:		_LOG _LNG _LNG _LNG _STR {
			if (NOT_HSTERR) {
				if (INPUTLOGNAME)
					verifylog($5);
				INPUTLOGINO = $2;
				INPUTLOGSIZE = $3;
				INPUTLOGPOSEND = $4;
			}
			free($5);
		}
		;

%%
