/*
   msg.c -- PIKT alert message routines

   $Header: /cvs/PIKT/pikt/src/piktd/msg.c,v 1.10 2002/08/23 17:00:24 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"

#include "piktd.h"

int newmailalert, newmailalarm, newprintalert, newprintalarm;

msg *mailmsg, *printmsg;
msg *mailmsgcurr, *printmsgcurr;

void regmsg(msg **, char *, int);

/* register alert header lines in the appropriate message queue
 */
void
althdr (msg **msgqueue, int type)
{
	struct tm *tm;
	time_t tnow;
	char altline[MAXLINE+1];
	char fmt[8];

	sprintf(altline, "\n%42s\n", "PIKT ALERT");
	regmsg(msgqueue, altline, type);
	tnow = time((time_t*)0);
	tm = localtime(&tnow);
	sprintf(altline, "%50s", asctime(tm));
	regmsg(msgqueue, altline, type);
	sprintf(fmt, "%%%ds\n", 37+(int)(strlen(hostname)/2));
	sprintf(altline, fmt, hostname);
	regmsg(msgqueue, altline, type);
	return;
}

/* register alarm header lines in the appropriate message queue
 */
void
alarmhdr (msg **msgqueue, alrm *alm, int type)
{
	char altline[MAXLINE+1];

	switch (alm->level) {
		case LOG_EMERG:
			sprintf(altline, "\n%s:\n", "EMERGENCY");
			break;
		case LOG_ALERT:
			sprintf(altline, "\n%s:\n", "URGENT");
			break;
		case LOG_CRIT:
			sprintf(altline, "\n%s:\n", "CRITICAL");
			break;
		case LOG_ERR:
			sprintf(altline, "\n%s:\n", "ERROR");
			break;
		default:
			sprintf(altline, "\n%s:\n", leveltypes[alm->level]);
			break;
	}
	regmsg(msgqueue, altline, type);
	sprintf(altline, "    %s\n", alm->name);
	regmsg(msgqueue, altline, type);
	requote(alm->task);
	sprintf(altline, "        %s\n\n", stripesc(alm->task));
	regmsg(msgqueue, altline, type);
	return;
}

/* format the current alert message string; register the string in the
   appropriate message queue
 */
void
altlin (msg **msgqueue, char *str, int type)
{
	char altline[MAXBUF+1];
	char *s, *lf;

	s = str;
	if ((*s == '\n') && ! *(s+1))	/* str is lf only */
		s++;
	else
		while (*s && (lf = strchr(s, '\n'))) {
			*lf = '\0';
			sprintf(altline, "%s%s\n", ! interactive ? "        " : nullstring, s);
			regmsg(msgqueue, altline, type);
			s = lf+1;
		}
	sprintf(altline, "%s%s\n", ! interactive ? "        " : nullstring, s);
	regmsg(msgqueue, altline, type);
	return;
}

/* queue the given message string to either the e-mail or print message
   queue
 */
void
regmsg(msg **msgqueue, char *line, int type)
{
	static char fnnam[] = "regmsg";
	msg *nmp;

	if (! interactive && (type != OUTPUT)) {
		nmp = (msg *) mymalloc(sizeof(msg), fnnam);
		nmp->next = (msg *) NULL;
		nmp->line = mystrdup(line, fnnam);
		if (! *msgqueue)
			*msgqueue = nmp;
		else {
			switch (type) {
				case MAIL:
					mailmsgcurr->next = nmp;
					break;
				case PRINT:
					printmsgcurr->next = nmp;
					break;
			}
		}
		switch (type) {
			case MAIL:
				mailmsgcurr = nmp;
				break;
			case PRINT:
				printmsgcurr = nmp;
				break;
		}
	} else {
		printf("%s", line);
		/* fflush(stdout); */
	}
	return;
}

/* dequeue the message lines to either an e-mail or print process
 */
void
mysendmsg (msg *messages, char *cmd)
{
	static char fnnam[] = "mysendmsg";
	msg *mp;
	FILE *altfile;

	if (messages) {
		if (cmd) {
			if ((altfile = popen(cmd, "w")) == NULL)
				logexit(PIKTLOG, LOG_ERR, ERRID_PROC_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, popen() error for %s", ALMNAM, cmd);
			foreach (mp, messages)
				fprintf(altfile, "%s", mp->line);
			pclose(altfile);
		} else
			logreturn(PIKTLOG, LOG_ERR, ERRID_MSG_CMD_NULL, __FILE__, __LINE__, fnnam, FALSE, "in %s, messages queued, but missing cmd", ALMNAM);
	}
	return;
}

/* dequeue/garbage collect messages from the indicated queue
 */
void
freemsg (msg **msgqueue)
{
	static msg *mp, *tmp;

	mp = *msgqueue;
	while (mp) {
		tmp = mp;
		mp = mp->next;
		FREE(tmp->line);
		free(tmp);
	}
	*msgqueue = (msg *) NULL;
	return;
}
